# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# Name:          characterCollector.py
# Purpose:       Character collector class.
#
#                CharacterCollector creates a MarkovChain
#                from a word stream (strings). The corresponding
#                Producer class is WordProducer which reverses the process.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright � 2021 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from common.collector import Collector
from common.textParser import TextParser
import pandas as pd

class CharacterCollector(Collector):

    def __init__(self, state_size=2, verbose=0, source=None, text=None, ignore_case=False):
        super().__init__(state_size, verbose, source,  domain='text')
        self.text = text
        self.ignore_case = ignore_case
        self.terminal_object = '~'
        self.initial_object = ' '

        self.countsFileName = '_charCounts' + '_0{}'.format(state_size)
        self.chainFileName = '_charsChain' + '_0{}'.format(state_size)
        
        self._text_parser = None

    def __str__(self):
        return f"CharacterCollector order={self.order} verbose={self.verbose} name={self.name} format={self.format}, source={self.source}, text={self.text}, ignoreCase={self._ignore_case}"
    
    def process(self, keys, word):
        keys += [ (''.join(word[i:i+self.order]), word[i+self.order])  for i in range(0, len(word)-self.order)]
        
    def _set_keys(self):
        #
        # create the keys 
        #
        prefix = self.initial_object
        self._initial_keys_df = self.counts_df[ [x.startswith(prefix) for x in self.counts_df['key']] ]
        self._initial_keys = set(self._initial_keys_df['key'].values.tolist())
        self._keys = set(self.counts_df['key'].values.tolist())
        if self.verbose > 0:
            print(f'total number of keys: {len(self._keys)}')
            print(f'number of initial keys: {len(self._initial_keys)}')

    def collect(self):
        """
        Run collection using the set parameters
        Returns MarkovChain result
        """
        word_keys = []
        if self.source is not None:
            self._text_parser = TextParser(source=self.source, ignore_case=self.ignore_case, maxlines=None, remove_stop_words=False)
            #
            # word counts are informational only
            #
            self.word_counts = self._text_parser.get_word_counts(sort_counts=True, reverse=True)
            self.words_df = self._text_parser.counts_df
            for w in self._text_parser.get_line_words():
                word = self.initial_object + w.strip() + self.terminal_object
                if len(word) >= self.order:
                    self.process(word_keys, word)
    
        elif self.text is not None:
            # add initial and terminal characters if needed
            s_rempunc = TextParser.remove_punctuation(self.text)
            words = s_rempunc.split(' ')
            for w in words:
                word = self.initial_object + w.strip() + self.terminal_object
                if len(word) >= self.order:
                    self.process(word_keys, word)
        
        #
        # create the counts_df DataFrame
        #          
        word_keys_df =  pd.DataFrame(data=word_keys, columns=['key','word'])
        words_ser = word_keys_df.value_counts(ascending=False)
        df = pd.DataFrame(words_ser, columns=['count'])
        self.counts_df = df.reset_index()
        if self.sort_chain:
            self.counts_df.sort_values(by=['key','word'], ignore_index=True, inplace=True)
        
        # create the MarkovChain from the counts by summing probabilities
        if self.counts_df.size > 0:
            self._set_keys()
            self._create_chain(initial=False)
        
        return self.markovChain
        
    def save(self):
        save_result = super().save()
        return save_result
    
if __name__ == '__main__':
    print(CharacterCollector.__doc__())




from common.ruleSet import RuleSet
from music.musicScale import MusicScale
from music.instruments import Instruments
from music21 import stream, instrument, key, meter
from music21 import note, tempo, metadata


class ScoreGen(object):
    
    def __init__(self, rule_set:RuleSet, resource_folder ="/Compile/dwbzen/resources/music", verbose=0, \
                 scale_name='Major', instrument_name='Soprano', key=key.Key('C'), title="Music Substiution" ):
        '''
            TODO - allow for more than 1 part (instrument). There needs to be a list of start notes,
            one for each part. instrument_name also needs to be a list.
        '''
        self.resource_folder = resource_folder
        self.scale_name = scale_name
        self.rule_set = rule_set
        self.verbose = verbose
        self.instrument_name = instrument_name
        #
        # create score metadata
        #
        self.tempo = tempo.MetronomeMark(number=100, referent=note.Note(type='quarter'))
        self.timeSignature = meter.TimeSignature('4/4')
        
        self.score = stream.Score()
        self.score.insert(0, metadata.Metadata())
        self.score.metadata.title = title
        
        self.part = stream.Part()    # the part created from pitch rules
        self.part.partName = instrument_name
        self.instruments = Instruments()
        self.instrument = self.instruments.instruments_pd.loc[instrument_name]
        clef = self.instruments.get_instrument_clef(instrument_name)
        self.part.insert(clef)
        self.part.insert(self.tempo)
        self.part.insert(instrument.Instrument(instrumentName=instrument_name))
        self.key = key
        self.part.insert(key)
        self.part.insert(self.timeSignature)
        
        self.state = dict()
        self.command_rules = rule_set.rules['commands']
        self.notes = []   # all the notes added

    def run(self, commands:[str], start_note=note.Note("C5", quarterLength=2)) -> stream.Score:
        '''The start_note must have a pitch with octave and a duration in quarterLengths
        
        ''' 
        root_note = note.Note(self.key.tonic)  # the scale root is determined by the key
        self.musicScale = MusicScale(resource_folder=self.resource_folder, scale_name=self.scale_name, root_note=root_note)
        self.scale = self.musicScale.scale
        self.start_note = start_note
        
        self.state['note'] = start_note
        self.state['previous_note'] = None
        self.state['musicScale'] = self.musicScale
        self.state['instruments'] = self.instruments
        self.state['instrument_name'] = self.instrument_name
        
        self.apply_commands(commands)
        self.score.append(self.part)
        return self.score

    def apply_commands(self, commands):
        '''Creates the Part notes from the commands string by executing the associated rules
            Sample commands: ['0/0.5', '+1/1.0', '-2/1.0', '+1/2.0', '+1/0.5', '0/duration', '0/1.0', '1/0.5']
        '''
        for command in commands:
            self.apply_command(command)
    
    def apply_command(self, command):
        if self.verbose > 0:
            print(f'Command: {command} -------------------------')
        group_dict = self.rule_set.splitter.match(command)
        if group_dict is None:
            return None
        interval = group_dict['interval']
        duration =  group_dict['duration']
        int_rule = self.command_rules['interval']
        dur_rule = self.command_rules['duration']
        int_rule(interval, self.state)    # apply to state['note']
        next_note = dur_rule(duration, self.state)    # apply to state['note']
        self.notes.append(next_note)
        self.part.append(next_note)

    



from music21 import note, duration
from music21.stream import Score
import copy

class MusicSubstitutionRules(object):
    
    trace = False
    
    @staticmethod
    def show(anote:note.Note):
        print(f"pitch: {anote.nameWithOctave}  duration: {anote.duration.quarterLength}")
    
    @staticmethod
    def score_postprocess(ascore:Score, instrument_name:str):
        '''Checks the pitch against the range of the given instrument, adjusting if necessary.
            TODO
        '''
        if MusicSubstitutionRules.trace:
            print('score_postprocess')
    
    @staticmethod
    def score_preprocess(ascore:Score):
        if MusicSubstitutionRules.trace:
            print('score_preprocess')
    
    @staticmethod
    def interval_rule(sd:str, state):
        anote = state['note']
        instrument_name = state['instrument_name']
        if sd == 'interval':
            scale_degrees = 0
        else:
            scale_degrees = int(sd)
            
        if MusicSubstitutionRules.trace:
            print(f' apply interval {scale_degrees} TO  {anote.nameWithOctave} {anote.duration.quarterLength}')
            
        musicScale = state['musicScale']   # music.MusicScale
        instruments = state['instruments']   # music.Instruments
        note_dur = anote.duration.quarterLength
        scale_note = musicScale.get_note(scale_degrees, anote)
        next_note = copy.deepcopy(scale_note)
        next_note.duration = duration.Duration(note_dur)
        state['previous_note'] = anote
        #
        # check the range of next_note and adjust if needed
        #
        if not instruments.is_in_range(instrument_name, next_note):
            instruments.adjust_to_range(instrument_name, next_note, inPlace=True)
        
        state['note'] = next_note
        #
        return next_note
    
    @staticmethod
    def duration_rule(d:str, state):
        '''Apply the duration command d to Note anote.
            Arguments:
                d - a float to multiply the Notes duration.quarterLength by
                anote - a Note
            Note if the resulting product is >4, it is reduced by half.
            If < 0.125, it is multiplied by 2
        '''
        anote = state['note']
        if d=='duration':
            multiplier = 1.0
        else:
            multiplier = float(d)
            
        if MusicSubstitutionRules.trace:
            print(f' apply duration {multiplier} TO {anote.nameWithOctave} {anote.duration.quarterLength}')
            
        old_dur = anote.duration.quarterLength
        new_dur = round(old_dur * multiplier, 3)
        if new_dur > 4.0:      # the maximum duration is a whole note
            new_dur = 0.5    
        elif new_dur < 0.125:  # the minimum duration is 32nd note
            new_dur = 2.0
            
        new_note = note.Note(anote.nameWithOctave, quarterLength=new_dur)
        if MusicSubstitutionRules.trace:
            print(f"\told_dur: {old_dur}  new_dur: {new_dur}\n")
        state['note'] = new_note
        return new_note
    
    if __name__ == '__main__':
        print("MusicSubstitutionRules")



# ------------------------------------------------------------------------------
# Name:          producer.py
# Purpose:       Runs CharacterCollector if needed and then WordProducer
#
#
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright 2022 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from music.musicProducer import MusicProducer
from music.noteCollector import NoteCollector
from music.intervalCollector import IntervalCollector
from music.instruments import Instruments
from common.utils import Utils
from common.markovChain import MarkovChain
from common.environment import Environment

import argparse
import pandas as pd
from music21 import key

class MusicProducerRunner(object):
    
    if __name__ == '__main__':
        """Produce a Score from MarkovChains
        
        Command line arguments:
            --chainFiles :  the name of existing MarkovChain files for intervals/notes and durations
                            By default it constructs the file names as: 
                            name + chainFileName of the associated collector type ('note' or 'interval') + '.json'
                            in the MusicCollector save_folder. For example, '--type interval --chainFiles bwv29_8' resolves
                            to 'bwv29_8_intervalsChain.json' and 'bwv29_8_durationsChain.json'
                            in the "/Compile/dwbzen/resources/music" folder (the default value of save_folder)
            --initial : The MusicCollectors specify an initial_object and a terminal_object. These values
                        are inserted automatically by the collector at the start of a collection unit,
                        in this case a Part. So for example, if the first interval in a Part is 2 (semitones)
                        the MarkovChain state will be interval.Interval(0) + interval.Interval(2)
                        If --initial is True, then when selecting a seed, pick one in the MarkovChain state space
                        that has the initial_object as the first element.
            --seed :    Optional initial seed.  The seed format depends on the source - Notes or Intervals
                        For intervals, specify semitones, for example --seed "5,-2"
                        For notes, specify the note with the octave, for example --seed "C#4,B3"
                        If unspecified, MusicProducer will select one randomly from the MarkovChain
                        depending on the value of --initial
     
        """
        
        parser = argparse.ArgumentParser(description="Produce a Music21 Score from a new or existing MusicCollector MarkovChain")

        #
        # Collector arguments
        #
        parser.add_argument("--order", "-o", help="the order of the Markov Chain", type=int, choices=range(1,5))
        parser.add_argument("--source", "-s", help="input file (.mxl or .musicxml), folder name, or composer name ('bach' for example)", default=None)
        parser.add_argument("--type", "-t", help="Source object type: notes or intervals", type=str, choices=['notes','intervals'], default='intervals')
        parser.add_argument("--parts", "-p", help="part name(s) or number(s) to include in building the MarkovChain", type=str, default=None)
        parser.add_argument("--filter", help="Apply filter to parts", type=str, default=None)
        #
        # mode only applies to 'notes' type
        #
        parser.add_argument("--mode", \
            help="Notes collection mode: ap (absolute pitch), dp (diatonic pitch), apc (absolute pitch class), dpc (diatonic pitch class) sd (scale degree)", \
                            type=str, choices=['ap','dp', 'apc','dpc', 'sd'], default='dpc')
        parser.add_argument("--format", "-f",help="Save output format. Default is json", type=str, choices=['csv','json','xlsx'], default='json' )
        parser.add_argument("--sort", help="Sort resulting MarkovChain ascending on both axes", action="store_true", default=False)
        #
        # Producer arguments
        #
        parser.add_argument("--chainFiles", "-c", help="Existing serialized MarkovChain and Durations files (--format type).",  type=str, default=None)
        
        parser.add_argument("--show", help="How to display resulting score", type=str, choices=['text','musicxml','midi'], default='musicxml')
        parser.add_argument("--name", help="Base name of generated files", type=str, default="MyScore")
        parser.add_argument("--title", help="Title of the generated Score", type=str, default="MyScore")
        parser.add_argument("--measures", "-m",  help="Total number of measures per part to produce",  type=int, default=10)
        parser.add_argument("--verbose", "-v", help="increase output verbosity", action="count", default=0)
        
        parser.add_argument("--produce", help="Comma-delimited list of Part name(s) to produce. A part name must also be a valid instrument name", \
                            type=str, action="extend", nargs="+", choices=Instruments.instrument_names,  default=None)
        parser.add_argument("--notes", help="For Interval and ScaleDegree chains, the starting note for each Part", type=str, default=None)
        parser.add_argument("--enforceRange", "-e", help="Enforce ranges of selected instruments", action="store_true", default=False)
        parser.add_argument("--key", "-k", \
            help="Key to use for all parts. Specify a single pitch. Lower case is minor, upper case is major. Key is adjusted for transposing instruments.", default=None)
        
        parser.add_argument("--seedNotes", help="Initial Notes seed. Number of args must be equal to the order of source chain", type=str, nargs='*', default=None)
        parser.add_argument("--seedIntervals", help="Initial Intervals seed. Number of args must be equal to the order of source chain", type=int, nargs='*', default=None)
        parser.add_argument("--seedScaleDegrees", help="Initial scaleDegrees seed. Number of args must be equal to the order of source chain", type=str, nargs='*', default=None)
        
        parser.add_argument("--initial", help="choose initial seed only", action="store_true", default=False)
        parser.add_argument("--recycle", help="How often to pick a new seed in terms of number of notes/intervals.", type=int, default=5)
        parser.add_argument("--trace", help="Show notes/intervals as they are produced", action="store_true", default=False)
        parser.add_argument("--duration","-d", help="Specify fixed duration as a quarterLength.", type=float, default=None)
        args = parser.parse_args()
        markovChain = None
        if args.verbose > 0:
            print('run ScoreProducer')
            print(args)
        randomseed = Utils.time_since()
        durationsChain = None
        markovChain = None
        collector = None
        if args.chainFiles is None and args.source is not None:
            #
            # run the appropriate collector first to produce the MarkovChains
            #
            if args.type== 'notes':
                collector = NoteCollector(state_size = args.order, verbose=args.verbose, source=args.source, parts=args.parts, collection_mode=args.mode )
            elif args.type == 'intervals':
                collector = IntervalCollector(state_size = args.order, verbose=args.verbose, source=args.source, parts=args.parts )
            
            collector.name = args.name
            collector.format = args.format
            collector.sort_chain = args.sort
            collector.set_score_filter(args.filter)    # could be None
            #
            # only for Bach works in the music21 corpus
            #
            collector.key_partName = 'Soprano'
            
            run_results = collector.run()
            if run_results['save_result']:
                print(f"MarkovChain written to file: {collector.filename}")
            markovChain = collector.markovChain
            durationsChain = collector.durationCollector.markovChain
        else:
            env = Environment.get_environment()
            save_folder = env.get_resource_folder('music')
            
            # use serialized MarkovChain in specified format, and Durations files in csv format as input
            file_list = []
            order = f'0{args.order}'
            file_list.append('{}/{}_{}Counts_{}.{}'.format(save_folder, args.chainFiles, 'durations', order, args.format))
            file_list.append('{}/{}_{}Chain_{}.{}'.format(save_folder, args.chainFiles, 'durations', order, args.format))
            if args.type == 'notes':
                file_list.append('{}/{}_{}Counts_{}_{}.{}'.format(save_folder, args.chainFiles, args.type, args.mode, order, args.format))
                file_list.append('{}/{}_{}Chain_{}_{}.{}'.format(save_folder, args.chainFiles, args.type, args.mode, order, args.format))
            else:   # there is no mode for intervals, as in 'bwv37x_intervalsChain_02.csv'
                file_list.append('{}/{}_{}Counts_{}.{}'.format(save_folder, args.chainFiles, args.type, order, args.format))
                file_list.append('{}/{}_{}Chain_{}.{}'.format(save_folder, args.chainFiles, args.type, order, args.format))                
            
            i = 0
            counts_df = None
            durations_counts_df = None
            for chainFile in file_list:
                file_info = Utils.get_file_info(chainFile)
                thepath = file_info["path_text"]
                ext = file_info['extension'].lower()
                if args.verbose > 0:
                    print(file_info)
                if not file_info['Path'].exists():
                    print(f"{thepath} does not exist")
                    exit()
                else:
                    if ext == 'json':
                        mc_df = pd.read_json(thepath, orient="index")
                    elif ext == 'csv':
                        mc_df = pd.read_csv(thepath)
                        # need to reindex and then drop the KEY column
                        new_index = mc_df['KEY']
                        mc_df.index = new_index.values
                        mc_df.drop(['KEY'],axis=1,inplace=True)
                        
                    if i==0:  # durations counts file
                        durations_counts_df = mc_df
                        i=1
                        continue
                    if i==1:
                        durationsChain = MarkovChain(args.order, durations_counts_df, chain_df=mc_df)
                        i=2
                    if i==2:    # notes/intervals counts file
                        counts_df = mc_df
                        i=3
                        continue
                    if i==3:
                        markovChain = MarkovChain(args.order, counts_df, chain_df=mc_df)

        
        musicProducer = MusicProducer(
            args.order, markovChain, durationsChain, \
            args.source, args.parts, args.produce, args.mode, \
            num=args.measures, verbose=args.verbose, \
            rand_seed=randomseed, producerType=args.type )
        
        musicProducer.show = args.show
        musicProducer.name = args.name
        musicProducer.recycle_seed_count = args.recycle
        musicProducer.initial = args.initial
        musicProducer.enforceRange = args.enforceRange
        musicProducer.trace_mode = args.trace
        musicProducer.fixed_duration = args.duration
        if args.key is not None:
            musicProducer.score_key = key.Key(args.key)
        #
        # order 2 examples:
        #  --seedIntervals -1 -2   for intervals
        #  --seedNotes C4 D4   for notes (dp, ap) --seedNotes C D   for notes (dpc, apc)
        #  --seedScaleDegrees 6 #5   for notes, collectionMode of 'sd'
        if args.seedIntervals is not None:
            musicProducer.set_seed(args.seedIntervals, args.type)
        if args.seedNotes is not None:
            musicProducer.set_seed(args.seedNotes, args.type)
        if args.seedScaleDegrees is not None:
            musicProducer.set_seed(args.seedScaleDegrees, args.type)
        
        #
        # intervals and notes (sd collectionMode) producerTypes need starting note(s) for each Part
        #
        if args.type == 'intervals' or args.mode == 'sd':
            musicProducer.add_part_notes(args.notes)
                
        theScore = musicProducer.produce()
        if theScore is not None:
            theScore.metadata.title = args.title
            theScore.show(args.show)
        print("Done")



# -*- coding: utf-8 -*- 
import pandas as pd
import string
import argparse
from markovify import split_into_sentences
from nltk.corpus import stopwords
from nltk import word_tokenize

class TextParser(object):
    """TextParser is a general purpose text parsing class.
    
    In addition to parsing blocks of text, it also compiles word statistics.
    
    Dependencies: This uses NLTK corpus for stop words, and markovify for splitting lines into sentences.
        pip install nltk markovify pandas
        
    TODO Known issues:
        * support treating double quote and sentence terminators (.!?)  as a words instead of punctuation.
        * remove stop words from sentences if remove_stop_words is True. Implement a staticmethod for this.
        
    """
    _punct = None
    _sw = stopwords.words('english') # Show some stop words - common words
    _utf_punct = '\N{LEFT SINGLE QUOTATION MARK}\N{RIGHT SINGLE QUOTATION MARK}\N{LEFT DOUBLE QUOTATION MARK}\N{RIGHT DOUBLE QUOTATION MARK}'
    _utf_quotes = '\N{LEFT DOUBLE QUOTATION MARK}\N{RIGHT DOUBLE QUOTATION MARK}'
    _quotes = '"' + _utf_quotes
    
    _prefix = ' '  # prefix for the initial word of a sentence or line

    def __init__(self, txt = None, source = None, maxlines=None, ignore_case=True, remove_stop_words=False):
        self._words = []        # words in order of appearance. Does not include stop words if remove_stop_words is True
        self._all_words = []    # words in order of appearance including stop words
        self._sentence_words = []    # words in sentence order with initial word of each sentence prefixed with a space
        self._line_words = None        # words in a line of text (which may be multiple sentences) with initial word of each prefixed with a space
        self._lines = []
        self._sentences = []
        self._word_set = None
        self._nlines = 0
        self._word_counts = {}
        self._maxlines = maxlines
        self.counts_df = None
        self.verbose = 0
        self.text = txt
        self._ignore_case = ignore_case    # if True convert all words to lower case
        self.source = source
        self._remove_stop_words = remove_stop_words
        self.replace_utf_punctuation = True      # replace UTF-8 single, double quotation marks with ' and " respectively
        self.treat_dblquotes_as_words = True    # if true treat " as a word (instead of a quote)
        self.initial_word_prefix = TextParser._prefix

        if source is not None:
            fp = open(source, "r")
            self.text = fp.read()
            self.text = self.text.replace('\t',' ')
        if self.text is not None and len(self.text) > 0:
            self.parse_text(self.text, maxlines=maxlines)
    
    @staticmethod
    def remove_punctuation(txt):
        if TextParser._punct is None:
            TextParser._punct =  string.punctuation.replace("'", "")
            TextParser._punct = TextParser._punct.replace('-', '')
            TextParser._punct = TextParser._punct + TextParser._utf_punct
            
        nopunc = [char for char in txt if char not in TextParser._punct]
        # Join the characters again to form the string.
        return ''.join(nopunc)
    
    @staticmethod
    def remove_quotes(txt):
        """Remove only the double quotes from a text string.
        
        This retains the final punctuation - period, ? or !
        """
        nopunc = [char for char in txt if char not in TextParser._quotes]
        # Join the characters again to form the string.
        return ''.join(nopunc)
    
    
    @staticmethod
    def replace_utf8_punctuation(txt):
        new_text = txt.replace('\N{LEFT DOUBLE QUOTATION MARK}', '"')
        new_text = new_text.replace('\N{RIGHT DOUBLE QUOTATION MARK}', '"')
        new_text = new_text.replace('\N{LEFT SINGLE QUOTATION MARK}', "'")
        new_text = new_text.replace('\N{RIGHT SINGLE QUOTATION MARK}', "'")
        return new_text
    
    @staticmethod
    def tokenize(self, txt) -> list:
        """Invokes NLTK word_tokenize function on a text string.
        
        Returns:
            a list of tokens which will include punctuation as individual list items
        
        """
        return word_tokenize(txt)
    
    def get_words(self):
        return self._words
    
    def get_all_words(self):
        return self._all_words
    
    def get_sentence_words(self):
        return self._sentence_words
    
    def get_line_words(self):
        """ Get words in a line of text with initial word of each prefixed with a space.
        
        A line will consist of one or more sentences.
        NOTE that stop words are not removed from line words.
        
        """
        if self._line_words is None:
            self._line_words = []
            for line in self._lines:
                s_rempunc = TextParser.remove_punctuation(line)
                words = s_rempunc.split(' ')
                line_words = []
                if self._ignore_case:
                    line_words = [str.lower(w) for w in words if len(w) > 0]
                else:
                    line_words = [w for w in words if len(w) > 0]
                    
                if len(line_words) > 0:
                    line_words[0] = self.initial_word_prefix + line_words[0]
                    self._line_words += line_words          
                        
        return self._line_words
    
    def get_lines(self):
        return self._lines
    
    def get_sentences(self):
        return self._sentences
    
    def size(self):
        return self._nlines
    
    def get_word_set(self):
        if self._word_set is None:
            self._set_word_set()
        return self._word_set
    
    def _set_word_set(self):
        self._word_set = set(self._words)
    
    def _set_word_counts(self, sort_counts=False, reverse=False):
        if len(self._word_counts) == 0:
            for w in self._word_set:
                self._word_counts |= {w:self._words.count(w)}
        if sort_counts:
            self._word_counts = dict(sorted(self._word_counts.items(), key=lambda item: item[1], reverse=reverse))
        self.counts_df = pd.DataFrame(data=self._word_counts.items(), columns=['word','count'])
    
    def get_word_counts(self, sort_counts=False, reverse=False):
        if self._word_set is None:
            self.get_word_set()
        self._set_word_counts(sort_counts, reverse)
        return self._word_counts
    
    def parse_text(self, txt, maxlines=None):
        if self.replace_utf_punctuation:
            self.text = TextParser.replace_utf8_punctuation(txt)
        for l in self.text.splitlines():
            if maxlines is not None and self._nlines >= maxlines:
                break
            if self.verbose > 0:
                print(f'line {self.nlines}: {l}')
            ls = split_into_sentences(l)
            for s in ls:
                s_rempunc = TextParser.remove_punctuation(s)
                words = s_rempunc.split(' ')
                allwords = [str.lower(w) for w in words if len(w) > 0]
                self._all_words += allwords
                sentence_words = []
                if self._remove_stop_words:
                    if self._ignore_case:
                        sentence_words = [str.lower(w) for w in words if len(w) > 0 and w.lower() not in TextParser._sw ]
                    else:
                        sentence_words = [w for w in words if len(w) > 0 and w.lower() not in TextParser._sw ]
                else:
                    if self._ignore_case:
                        sentence_words = [str.lower(w) for w in words if len(w) > 0]
                    else:
                        sentence_words = [w for w in words if len(w) > 0]
                self._words += sentence_words
                
                if len(sentence_words) > 0:
                    sentence_words[0] = self.initial_word_prefix + sentence_words[0]
                    self._sentence_words += sentence_words
                    
                self._sentences.append(s)
            self._lines.append(l)
            self._nlines += 1
        self._set_word_set()
        return self._nlines

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-t", "--text", help="in-line text input. One of --text or --source must be specified", default=None)
    parser.add_argument("-s", "--source", help="input file name")
    parser.add_argument("-v","--verbose", help="increase output verbosity", action="count", default=0)
    parser.add_argument("-i","--ignoreCase", help="ignore input case", action="store_true", default=False)
    parser.add_argument("-r", "--remove_stop_words", help="remove common stop words", action="store_true", default=False)
    parser.add_argument("-l", "--lines", help="Maximum number of lines to read. If not set, reads all source lines", type=int, default=None)
    args = parser.parse_args()
    text_parser = TextParser(txt=args.text, source=args.source, maxlines=args.lines, remove_stop_words=args.remove_stop_words)
    word_counts = text_parser.get_word_counts(sort_counts=True, reverse=True)
    counts_df = text_parser.counts_df
    #
    # display the top 20
    print(counts_df.head(20))
    print(word_counts)
    
    sentences = text_parser.get_sentences()
    print(f'{len(sentences)} sentences\n ')
    #for s in sentences: print(s)



# ------------------------------------------------------------------------------
# Name:          collectorProducer.py
# Purpose:       Base class for Collector and Producer.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2021 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from common.utils import Utils
from common.environment import Environment
import sys
import pandas as pd

class CollectorProducer(object):
    
    def __init__(self, state_size=2, verbose=0, source=None, domain='text'):
        
        self.order = state_size
        self.domain = domain
        self.verbose = verbose
        self.source = source    # file input or DataFrame source
        self.name = None        # set by the user on the command line
        self.format = None      # csv or json
        
        self.chain_filename = None          # MarkovChain.chain_df filename
        self.counts_filename = None         # MarkovChain.counts_df filename
        
        self.chain_df = None        # pd.DataFrame()
        self.counts_df = None       # pd.DataFrame()
        #
        # update Environment to reflect your environment
        #
        env = Environment.get_environment()
        self.save_folder = env.get_data_folder(domain)   # for example "/Compile/dwbzen/data"
        
        self.chainFileName = '_chain'       # appended to self.name for the MarkovChain file
        self.countsFileName = '_counts'     # appended to self.name for the counts file
        
        
    def __repr__(self):
        return f"CollectorProducer {self.order}"

    @staticmethod
    def save_df(df:pd.DataFrame, filename, file_format, orient='index', sheet_name='Sheet 1'):
        result = True
        if file_format=='csv':
            df.to_csv(filename)
        elif file_format=='json':
            dumped = Utils.get_json_output(df, orient)
            with open(filename, 'w') as f:
                f.write(str(dumped))
        elif file_format=='xlsx' or file_format=='excel':
            df.to_excel(filename, sheet_name=sheet_name, index=False)
        else:
            result = False
            print("Empty DataFrame(s)", sys.stderr)
        return result
    
    def save(self):
        """Saves the MarkovChain and counts DataFrame files
        
        The chain_df DataFrame of the MarkovChain is saved in the specified format,
        csv, Excel or json, to the self.save_folder directory.
        The filename is the name of the MarkovChain (self.name) + the chainFileName.
        The chainFileName default is '_chain' and this is typically set
        to something more appropriate in Collector subclasses.
        For example, '_charsChain'.
        
        Similarly, the counts_df DataFrame is saved using the self.countsFileName
        to create the save filename. For example, '_charsCounts'
        
        """
        save_result = False
        if self.name is not None:   # format will always be set to something, even if name is None
            self.markovChain.name = self.name
            
            self.filename = "{}/{}{}.{}".format(self.save_folder, self.name, self.chainFileName, self.format)
            self.counts_file = "{}/{}{}.{}".format(self.save_folder, self.name, self.countsFileName, self.format)
            if self.verbose > 1:
                print(f"output filenames '{self.filename}' counts: {self.counts_file}")
            save_result = CollectorProducer.save_df(self.chain_df, self.filename, self.format)
            save_result = save_result and CollectorProducer.save_df(self.counts_df, self.counts_file, self.format)
            
        return save_result
    


'''
Created on Apr 10, 2021

@author: don_bacon
'''
import fileinput, taglib, argparse, re, os
from sys import __stderr__

class Path_titles(object):
    
    def __init__(self, file_name='', matchPath=''):
        '''
        Constructor
        '''
        self.filename = file_name
        self.count = 0
        self.renamedCount = 0
        self.verbose = 0
        self.list=False
        self.errors = []
        self.noArtist = []
        self.matchPatterns = []
        self.matchPath = matchPath
        self.rename = False     # if true, replace the matching text with the ARTIST name in the filename
        self.delete = False     # if true, delete the matching text from the filename and rename
        self.verify = False
        
    def save_match_patterns(self, patterns:list):
        '''
        patterns is a list of compiled regular expressions
        '''
        for pattern in patterns:
            self.matchPatterns.append(re.compile(pattern, re.IGNORECASE))
    
    def process_line(self, line: str) -> tuple:
        parts = line.rsplit('/')
        n = len(parts)
        filename = parts[n-1]
        path = '/'.join(parts[0:n-1])
        if(path.startswith('./')):
            path = path.lstrip('.')
        return (path, filename)
    
    def rename_file(self, oldfile_path:str,  newfile_path:str) -> bool:
        try:
            os.rename(oldfile_path, newfile_path)
            return True
        except OSError as err:
            self.errors.append(oldfile_path)
            print('OSError: Could not rename file: "{}"  because \n\t{}'.format(oldfile_path, err.strerror), file=__stderr__)
            return False
    
    def process_file(self):
        audioFilename = ''
        with fileinput.input(files=self.filename) as f:
            for line in f:
                if(self.verbose >= 2):
                    print('line: {}'.format(line))
                (path, filename) = self.process_line(line)
                if(self.matchPath is not None and path.find(self.matchPath) < 0):
                    continue
                filename=filename.rsplit('\n')
                if(self.verbose > 2):
                    print("{p}\t{f}".format(p=path, f=filename))
                try:
                    audioFilename = filename[0]
                    filepath =  path + "/" + audioFilename
                    song=taglib.File(filepath)
                except OSError:
                    self.errors.append(filepath)
                    print('Could not read file: {}'.format(filepath), file=__stderr__)
                    continue
                if('ARTIST' in song.tags):
                    artists = song.tags['ARTIST']   # returns a list of artists delimited by slash "/"
                    song.close()
                    artist = ','.join(artists)
                    artist = artist.replace('/', ',')       # replace '/' delimiter with a comma
                    if(self.list):
                        print('{}\t{}'.format(artist, filepath))
                    if(self.rename or self.delete):
                        # rename the file, replacing the matched portion of the filename with the ARTIST tag 
                        # if delete is true, delete the matching portion from the filename
                        #
                        for regx in self.matchPatterns:
                            matchit = regx.match(audioFilename)
                            if(matchit is not None):
                                if(self.rename):
                                    new_audioFilename = '{} - {}'.format(artist, audioFilename[matchit.end()+1:])
                                elif(self.delete):
                                    new_audioFilename = audioFilename[matchit.end():]
                                new_audioFile_path = '{}/{}'.format(path, new_audioFilename)
                                if(self.verbose >0):
                                    print('\t match on "{}" start,end {}, {}'.format(audioFilename,matchit.start(), matchit.end()))
                                if(self.verify):
                                    answer = input('\t Rename "{}" to "{}" ? (y/n/q): '.format(audioFilename, new_audioFilename)).rstrip().lower()
                                    if 'y' == answer:
                                        renamed = self.rename_file(filepath, new_audioFile_path)
                                        if renamed:
                                            print("\t file renamed to {}".format(new_audioFilename))
                                            self.renamedCount = self.renamedCount + 1
                                        else:
                                            print("\t unable to rename to {}".format(new_audioFilename), file=__stderr__)
                                    elif 'n' == answer:
                                        print('{} not renamed'.format(audioFilename))
                                    elif 'q' == answer:
                                        return
                                else:
                                    renamed = self.rename_file(filepath, new_audioFile_path)
                                    if renamed:
                                        print("\t file renamed to {}".format(new_audioFilename))
                                        self.renamedCount = self.renamedCount + 1
                                    else:
                                        print("\t unable to rename to {}".format(new_audioFilename), file=__stderr__)
                else:
                    self.noArtist.append(filepath)
                    song.close()
                self.count = self.count + 1
        return self.count


    def getArtist(self):
        pass
 
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Process audio files.')
    parser.add_argument('file',  help='text file of audio file path(s).')
    parser.add_argument('--path', help='audio files match path.')
    parser.add_argument('--list', '-l', action='store_true', help='print artists and titles of songs')
    parser.add_argument('--rename', action='store_true', help='rename file if name matches one of the patterns')
    parser.add_argument('--delete', action='store_true', help='delete the matching string from the file name matches one of the patterns')
    parser.add_argument('--verify', action='store_true', help='verify file rename')    
    parser.add_argument('--patterns', default='\d\d-\d\d-', help='filename matching pattern regular expression(s)' )
    parser.add_argument('--verbose', '-v', action='count', default=0, help='verbose output level')
    
    args = parser.parse_args() 
    filename=args.file
    path = args.path
    if(args.delete and args.rename):
        print("You can't both rename the file and delete the prefix", end='\n', file=__stderr__)
        exit()

    pathTitles = Path_titles(filename)
    pathTitles.matchPath = path
    pathTitles.verbose = args.verbose
    pathTitles.list = args.list
    pathTitles.rename = args.rename
    pathTitles.delete = args.delete
    pathTitles.verify = args.verify
    pathTitles.save_match_patterns(args.patterns.split(','))
    count = pathTitles.process_file()
    print("number of titles: {}".format(count))
    print('{} files renamed'.format(pathTitles.renamedCount))
    if(len(pathTitles.noArtist) > 0):
        print('the following {} titles have no artist'.format(len(pathTitles)))
        for s in pathTitles.noArtist:
            print(s)
    
    



# ------------------------------------------------------------------------------
# Name:          noteCollector.py
# Purpose:       Base music collector class. Subclasses: NoteCollector, IntervalCollector
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2022 Donald Bacon
#
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from common.collector import Collector
from music.instruments import Instruments
import music
import pandas as pd

class MusicCollector(Collector):
    
    def __init__(self, state_size=2, verbose=0, source=None, parts=None):
        super().__init__(state_size, verbose, source, domain='music')
        
        self.instruments = Instruments(self.verbose)

        self.corpus_folder="/Compile/music21/music21/corpus"    # the default corpus folder
        self.terminal_object = None     # set in derived classes
        self.initial_object = None      # set in derived classes
        self.score = None       # if source is a single Score
        self.scores = []        # a list of Score if collecting from a corpus
        self.titles = []        # the titles of all Score(s)
        self.part_names = []    # optional part names used to filter parts of score(s)
        self.part_numbers = []  # optional part numbers used to filter parts
        self.parts = None       # comma-delimited part names from the command line
        self.score_partNumbers = set()     # the part numbers extracted from the score or scores
        self.score_partNames = set()       # the part names extracted from the score or scores
        self.durationCollector = None
        self.durations_df = None
        if parts is not None:
            self.add_parts(parts)
        self.enforceRange = True
        self.counts_df = None
        self.chain_df = None
        #
        # filter string to apply against Part(s)
        # format is: property=value, for example "mode=minor"
        # The filter is applied after the score_key is set (if it's not None)
        #
        self._filter = None
        self.score_filters = {}     # mode filter can be 'major' or 'minor'
        #
        # if key_partName is not None, all the Key for score parts is
        # set to the Key of the named part.
        # Essentially this becomes the Key to use for an entire Score
        #
        self.key_partName = None
    
        
    def set_score_filter(self, filter_string):
        if filter_string is not None and len(filter_string) >= 3 and '=' in filter_string:
            self._filter = filter_string
            temp = filter_string.split('=')
            if temp[0] == 'mode':
                self.score_filters['mode'] = temp[1]
            
    def add_parts(self, parts) -> str:
        if parts is not None:
            self.parts = parts
            parts_list = parts.split(",")   # could be numbers or names
            for p in parts_list:
                if p.isdigit():   # all digits
                    self.part_numbers.append(int(p))
                else:
                    self.part_names.append(p)               


    def collect(self):  # implement in derived classes
        return None
    
    def collect_durations(self, source:pd.DataFrame ):
        self.durationCollector = music.DurationCollector(self.order, self.verbose, source, self.parts)
        self.durationCollector.score = self.score
        if self.name is not None and len(self.name) > 0:
            self.durationCollector.name = self.name
        self.durationCollector.format = self.format
        self.durationCollector.sort_chain = self.sort_chain
        run_results = self.durationCollector.run()
        self.durations_df = self.durationCollector.durations_df
        return run_results
    
    def save(self):
        save_result = super().save()
        return save_result
        



from threading import Lock

class Environment(object):
    _environ = None
    _lock = Lock()
    
    def __init__(self, package_name):
        """Initialize the running Environment by setting global environment variables
        
        TODO: implement using a properties config file
        
        """
        self.name = package_name
        self.resource_base = '/Compile/dwbzen/resources'    # for managed resources
        self.data_base = '/Compile/dwbzen/data'             # for output files
        self.package_name = package_name
        self.resources = {}
        self.data = {}
        self.resources['music'] = self.resource_base + '/music'
        self.resources['text'] = self.resource_base + '/text'
        self.data['music'] = self.data_base + '/music'
        self.data['text'] = self.data_base + '/text'
        self._items = {}
        
    def __repr__(self):
        return '<Environment>'
        
    def get_resource_folder(self, package):
        if package in self.resources:
            return self.resources[package]
        else:
            return self.resource_base
    
    def get_data_folder(self, package):
        if package in self.resources:
            return self.data[package]
        else:
            return self.data_base
        
    def add_item(self, key, value):
        self.items[key] = value
    
    @property
    def items(self):
        return self._items
    
    def get_item(self, key):
        return self.items[key]

    @staticmethod
    def get_environment():
        env = None
        Environment._lock.acquire()
        if Environment._environ is not None:
            env = Environment._environ
        else:
            Environment._environ = Environment('dwbzen')
            env = Environment._environ
        Environment._lock.release()
        return env




'''
Created on Feb 18, 2021

@author: don_bacon
'''
import json
import pandas as pd

class Song(object):
    '''
    Represents a JSON formatted song
    '''

    def __init__(self,  raw_data={}, json_file_name=''):
        '''
        Constructor
        '''
        self.song_notes = []
        self.song_harmony = []
        self.section_notes = {}   # key is section name
        self.section_harmony = {} # key is section name
        self.section_names = []
        self.data = {}
        self.song_name = 'None'
        if (len(json_file_name) >0):
            _song_file = json_file_name
            with open(_song_file, "r") as read_file:
                self.data = json.load(read_file)
        else:
            self.data = raw_data
            self._get_song_data()
    
    def _get_song_data(self):
        if('name' in self.data):
            self.song_name = self.data['name']
            self.score_key = self.data['scoreKey']
        for section in self.data['sections']:
            section_name = section['name']
            self.section_names.append(section_name)
            print('getting section name: {}'.format(section_name))  
            measures = section['measures']
            self.section_notes[section_name] = []
            self.section_harmony[section_name] = []
            for measure in measures:
                measure_number = measure['number']
                if('timeSignature' in measure):
                    self.time_signature = measure['timeSignature']
                if('melody' in measure):
                    melody = measure['melody']
                notes = melody['notes']
                if(len(notes) > 0):
                    for n in notes:
                        n['section']=section_name
                        n['measure'] = measure_number
                        self.song_notes.append(n)
                        self.section_notes[section_name].append(n)
                if('harmony' in measure):
                    harmony = measure['harmony']
                    if(len(harmony) > 0):
                        for h in harmony:
                            h['section'] = section_name
                            h['measure'] = measure_number
                            self.section_harmony[section_name].append(h)
                            self.song_harmony.append(h)
        '''
        Create data frames for song_notes and song_harmony
        '''
        self.df_notes = pd.DataFrame(self.song_notes)
        self.df_notes = self.df_notes.fillna(value={'chord':'0'})
        self.df_harmony = pd.DataFrame(self.song_harmony)
        self.df_harmony = self.df_harmony.fillna(value={'chord':'0'})

    def __iter__(self):
        ''' Iterate over the song_notes '''
        return self.song_notes.__iter__()




#
# ------------------------------------------------------------------------------
# Name:          noteCollectorRunner.py
# Purpose:       Run NoteCollector
#
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2022 Donald Bacon

# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from common.collector import Collector
from music.noteCollector import NoteCollector
import argparse

class NoteCollectorRunner(Collector):
    """ Executes NoteRunner with the parameters provided in the command line
    
    """
    
    if __name__ == '__main__':
        parser = argparse.ArgumentParser()
        parser.add_argument("--order", "-o", help="the order of the Markov Chain", type=int, choices=range(1,5))
        parser.add_argument("-s", "--source", help="input file (.mxl or .musicxml), folder name, or composer name ('bach' for example)")
        parser.add_argument("-v","--verbose", help="increase output verbosity", action="count", default=0)
        parser.add_argument("-n","--name", help="Name of resulting MarkovChain, used to save to file", type=str)
        parser.add_argument("-f","--format", help="Save output format. Default is csv", type=str, choices=['csv','json','xlsx'], default='csv' )
        parser.add_argument("--sort", help="Sort resulting MarkovChain ascending on both axes", action="store_true", default=False)
        parser.add_argument("-d","--display", help="display resulting MarkovChain in json or cvs format",type=str, choices=['csv','json','chain'] )
        parser.add_argument("-p","--parts", help="part name(s) or number(s) to include in building the MarkovChain", type=str)
        parser.add_argument("-m", "--mode", \
            help="Notes collection mode: ap (absolute pitch), dp (diatonic pitch), apc (absolute pitch class), dpc (diatonic pitch class) sd (scale degree)", \
                            type=str, choices=['ap','dp', 'apc','dpc', 'sd' ], default='dpc')
        parser.add_argument("--enforceRange", "-e", help="Enforce ranges of selected instruments, applies to diatonic collection modes.", action="store_true", default=False)
        parser.add_argument("--show", help="Show the original or transposed score when collection is complete. Does not apply when source is multiple scores.", \
                            type=str, choices=['original', 'transposed', 'none'], default='none')
        parser.add_argument("--filter", help="Apply filter to parts", type=str, default=None)
        parser.add_argument("--keypart", help="Part that determines the overall score Key", type=str, default=None)

        args = parser.parse_args()
        if args.verbose > 0:
            print('run NoteCollector')
            print(args)

        collector = NoteCollector(state_size = args.order, verbose=args.verbose, source=args.source, parts=args.parts, \
                                        collection_mode=args.mode, enforce_range=args.enforceRange )
        collector.name = args.name
        collector.format = args.format
        collector.sort_chain = args.sort
        collector.set_score_filter(args.filter)    # could be None
        collector.key_partName = args.keypart      # could be None
        #
        # only for Bach works in the music21 corpus
        #
        # collector.key_partName = 'Soprano'
        
        if args.verbose > 0:
            print(collector.__str__())

        
        run_results = collector.run()
        if run_results['save_result']:
            print(f"MarkovChain written to file: {collector.filename}")
        if args.display is not None:
            # display in specified format
            if args.display=='json':
                print(collector.get_json_output())
            elif args.display=='csv':
                print(collector.markovChain.chain_df.to_csv(line_terminator='\n'))
            else:   # assume 'chain' and display the data frame
                print(collector.markovChain.chain_df)
                
        nscores = collector.number_of_scores
        if args.show != 'none' and nscores==1:
            theScore = collector.score
            if args.show == 'transposed':
                theScore = collector.transposed_score
            elif args.show == 'original':
                theScore = collector.score
                
            if theScore is not None:
                theScore.show()





#
# ------------------------------------------------------------------------------
# Name:          characterCollector.py
# Purpose:       Executes the CharacterCollector
#
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2021 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from common.collector import Collector
from common.characterCollector import CharacterCollector
import argparse

class CharacterCollectorRunner(Collector):
    """Executes the CharacterCollector class.
    
    
    """
    
    if __name__ == '__main__':
        parser = argparse.ArgumentParser()
        parser.add_argument("order", help="the order of the Markov Chain", type=int, choices=range(1,5))
        parser.add_argument("-t", "--text", help="in-line text input. One of --text or --source must be specified")
        parser.add_argument("-s", "--source", help="input file name")
        parser.add_argument("-v","--verbose", help="increase output verbosity", action="count", default=0)
        parser.add_argument("-i","--ignoreCase", help="ignore input case", action="store_true", default=False)
        parser.add_argument("-n","--name", help="Name of resulting MarkovChain, used to save to file", type=str)
        parser.add_argument("-f","--format", help="Save output format. Default is csv", type=str, choices=['csv','json','xlsx'], default='csv' )
        parser.add_argument("--sort", help="Sort resulting MarkovChain ascending on both axes", action="store_true", default=False)
        parser.add_argument("-d","--display", help="display resulting MarkovChain in json or cvs format",type=str, choices=['csv','json','chain'] )
        args = parser.parse_args()
        if args.verbose > 0:
            print('run CharacterCollector')
            print(args)
            
        collector = CharacterCollector(state_size = args.order, verbose=args.verbose, source=args.source, text=args.text, ignore_case=args.ignoreCase)
        collector.name = args.name
        collector.format = args.format
        collector.sort_chain = args.sort
        if args.verbose > 0:
            print(collector.__repr__())
        run_results = collector.run()
        if run_results['save_result']:
            print(f"MarkovChain written to file: {collector.filename}")
        if args.display is not None:
            # display in specified format
            counts_df = collector.counts_df
            print(f'counts_df:\n{counts_df.head(10)}\n')
            if args.display=='json':
                print(collector.get_json_output())
            elif args.display=='csv':
                print(collector.markovChain.chain_df.to_csv(line_terminator='\n'))
            else:   # assume 'chain' and display the data frame
                print(collector.markovChain.chain_df)



'''
Created on Sep 8, 2021

@author: don_bacon
'''

VERSION_TUPLE = (0, 1, 2)
__version__ = ".".join(map(str, VERSION_TUPLE))
__title__ = 'common'
__description__ = 'Python Common library'
__author__ = 'Donald Bacon'
__author_email__ = 'dwbzen@gmail.com'
__license__ = 'MIT License'
__copyright__ = 'Copyright 2022 Donald Bacon'



# ------------------------------------------------------------------------------
# Name:          collector.py
# Purpose:       Base collector class.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2021 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

import pandas as pd
from common.markovChain import MarkovChain
from common.utils import Utils
from common.collectorProducer import CollectorProducer

class Collector(CollectorProducer):
    """Collects data from a defined source and builds a MarkovChain of a given order.
    
        Collector subclasses are tailored to the specific data type being collected and how that data is sourced.
        A collector has a corresponding Producer class that takes
        a MarkovChain as input and produces data.
        For example, a CharacterCollector creates a MarkovChain
        from a word stream (strings). The WordProducer reverses the process.
        
        The terminal_object and initial_object must be set appropriately
        by derived (concrete) classes. 
        terminal_object should be set to a value that does not occur
        in the collection source data. For example, CharacterCollector
        sets the value to '~'. IntervalCollector sets the value to
        a interval.Interval to interval.Interval(99).
        The initial_object represents the start of whatever unit is being collected
        This is used to identify an initial seed
    """
    
    def __init__(self, state_size, verbose=0, source=None, domain='text'):
        """Initialize elements common to all Collector subclasses.
        
        The chain_df and counts_df DataFrames index and columns are set by the subclass 
        and is totally dependent on the target domain (for example 'text') and units (for example, characters or words)
        """
        super().__init__(state_size, verbose=verbose, source=source, domain=domain)
        
        self.stateSpace_type = None
        self.markovChain =  None

        self.sort_chain = True

        self.terminal_object = None         # must be set in derived classes
        self.initial_object = None          # must be set in derived classes
        
        self._initial_keys_df = None     # DataFrame created by collect() from counts_df
        self._initial_keys = None        # unique list of key values of initial_keys_df
        self._keys = None                # unique key values of counts_df
        
    def __repr__(self, *args, **kwargs):
        return "Collector"
    
     
    def get_initial_keys(self):
        return self._initial_keys
    
    def get_keys(self):
        return self._keys
           
           
    def _add_counts_to_model(self, model: dict, countsdf: pd.DataFrame, akey: str):
        key_df = countsdf[countsdf['key'] ==  akey]
        s = (key_df['count']/key_df['count'].sum()).cumsum()
        model[akey] = key_df.assign(prob=s )

            
    def _create_chain(self, initial=True) -> MarkovChain:
        """Creates a MarkovChain instance from a counts_df DataFrame
        
        The MarkovChain model is chain_df DataFrame with the columns: key, word, count (from count_df) and prob (the probability)
        'key' is a string of n-words/characters (n=order of the chain) that appear in the source,
        'word' is the next token, and 'prob' is the probability that combination appears in the source.
        The probabilities are computed from the counts.
        
        Parameters:
            initial - if True (the default) create the MarkovChain from initial_keys only
        """

        thekeys = self._initial_keys
        if not initial:
            thekeys = self._keys
        
        self.markovChain = MarkovChain(self.order, self.counts_df, keys=thekeys, chain_df=None,  myname=self.name)
        if self.sort_chain:
            self.chain_df = self.markovChain.chain_df.sort_values(by=['key','word'], ignore_index=True, inplace=False)
            self.markovChain.chain_df = self.chain_df
        
        return self.markovChain
    
    def run(self):
        """Invokes the derived class's collect() function and invokes save()
        
        Invokes the derived class's collect() function and if the resulting
        MarkovChain is valid, invokes save()
        Returns the boolean results of the collect() and save() as a dict()
        with keys 'collect_result' and 'save_result'
        """
        
        save_result = False
        collect_result = False
        self.collect()
        if self.verbose > 1:
            print(f" Counts:\n {self.counts_df}")
            print(f" MarkovChain:\n {self.markovChain}")
            if self.verbose > 2:
                print(self.markovChain.__repr__())        
        
        if self.markovChain is not None and len(self.markovChain.chain_df) > 0:
            collect_result = True
            save_result = self.save()
            
        return {'collect_result' : collect_result, 'save_result' : save_result}

    def collect(self):
        """
        Override in subclass
        """
        return None

    def save(self):
        save_result = super().save()
        return save_result
    
    def get_json_output(self):
        return Utils.get_json_output(self.chain_df)

if __name__ == '__main__':
    print(Collector.__doc__)
        




# ------------------------------------------------------------------------------
# Name:          noteCollector.py
# Purpose:       Note collector class.
#
#                NoteCollector creates a pair of MarkovChains
#                from a Note stream (strings) for the Notes and Durations.
#                The corresponding Producer class is PartProducer which reverses the process.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2022 Donald Bacon
#
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from music.musicCollector import MusicCollector
from music.musicUtils import MusicUtils
from common.markovChain import MarkovChain
from common.collectorProducer import CollectorProducer
import pandas as pd
from music21 import  converter, corpus, note

class NoteCollector(MusicCollector):
    """Collect Notes from one or more scores into a MarkovChain
    
    There are 6 collection modes: absolute, diatonic, absolute pitch class, diatonic pitch class,
    scale degree number (sdn), and scale degree roman (sdr).
    In absolute mode the note pitches used are exactly those that appear in the score(s) without alteration.
    In diatonic mode, the pitch used is the diatonic equivalent for the Key of C-major.
    For example consider the note "F#5". Relative to the key of D-major this is the 3rd step in the scale.
    Transposing to C-major this becomes "E5".
    The same note relative to the score key of Bb-major is #5, so G# in C-Major.
    
    When transposing minor keys, the natural minor scale is used (which is the same as melodic minor descending)
    
    Absolute pitch class collects only the pitch class of each note. 
    Pitch class is a music21 pitch attribute that is an integer number from 0 to 11 where 0=C, 1=C# etc.
    Diatonic pitch class collects the pitch class of each note transposed to C-Major.
    The octave portion of the note is lost when collecting pitch class. 
    However the MusicProducer provides ways to reinsert an octave (for each part) when generating scores.
    
    Scale degree collects the scale degree of each pitch relative to the Key of the Part.
    Scale degree records the appropriate scale degree as an integer, 1 to 7.
    Both are relative to the underlying Key or Scale. And both prepend an accidental if there is one.
    
    Note that the Mode of the Scale makes a difference when collecting scale degrees.
    For example, E-minor and G-major have the same key signature, but the scale degrees are different -
    an 'e' is 1 in E-minor, but 6 in G-major. This can be problematic in some of the musicxml scores
    that have parts in minor mode.
    For example the music21 Bach corpus scores consistently have the "real" key in the Soprano part,
    the other parts the relative major.
    To account for this the collector includes two additional features: filter and keypart.
    If set, the filter will return only scores that meet the filter criteria. 
    At present the only filter supported is 'mode' which can have a values 'major' or 'minor'.
    Keypart is the part name that determines the actual (real) key. 
    If set, then the key to all parts is set to the keypart key.
    So the appropriate command-line arguements would be: --mode sd --filter "mode=minor" --keypart Soprano
    
    """
    
    initial_object = note.Note('C0')
    terminal_object = note.Note('C#8')
        
    def __init__(self, state_size=2, verbose=0, source=None, parts=None, collection_mode='dpc', enforce_range=True):
        super().__init__(state_size, verbose, source, parts)
        
        self.initial_object, self.terminal_object = NoteCollector.initialize_initial_terminal_objects()
        
        self.countsFileName = '_notesCounts_' + collection_mode + '_0{}'.format(state_size)
        self.chainFileName = '_notesChain_' + collection_mode + '_0{}'.format(state_size)
        self.notes_df_fileName = '_notes_df'        # self.notes_df saved
        
        self.notes_df = pd.DataFrame()
        self.collection_mode = collection_mode      # default is absolute pitch
        self.enforce_range = enforce_range          # applies only to dp and dpc collection modes

        self.number_of_scores = 0                   # can be >1 if searching a corpus
        self.transposed_scores = []                 # if collection mode is diatonic pitch or pitch class
        self.transposed_score = None                # a single score if diatonic collection
        MusicUtils.verbose = verbose
        
        self.pitch_class_mode = (collection_mode == 'dpc' or collection_mode == 'apc')
        self.pitch_mode =  (collection_mode == 'dp' or collection_mode == 'ap')
        self.scale_degree_mode = collection_mode.startswith('sd')
        self.diatonic = collection_mode.startswith('d')
        self.absolute = (collection_mode.startswith('a') or collection_mode.startswith('s'))

        #
        # collection_scores will reference self.scores if collection mode is absolute pitch or pitch class
        # or self.transposed_scores if collection mode is diatonic
        #
        self.collection_scores = []
        self.source_path = source
            
    @staticmethod
    def initialize_initial_terminal_objects() -> (pd.DataFrame, pd.DataFrame):

        initial_dict = {'note':NoteCollector.initial_object, 'part_number':1, 'part_name':'note', \
               'nameWithOctave':NoteCollector.initial_object.nameWithOctave, 'pitch':str(NoteCollector.initial_object.pitch), \
               'duration':NoteCollector.initial_object.duration, \
               'scaleDegree':'1', 'ps':12.0, \
               'pitchClass':NoteCollector.initial_object.pitch.pitchClass, 'name':NoteCollector.initial_object.name}
        
        terminal_dict = {'note':NoteCollector.terminal_object, 'part_number':1, 'part_name':'note', \
               'nameWithOctave':NoteCollector.terminal_object.nameWithOctave, 'pitch':str(NoteCollector.terminal_object.pitch), \
               'duration':NoteCollector.terminal_object.duration, \
               'scaleDegree':'#1', 'ps':109.0, \
               'pitchClass':NoteCollector.terminal_object.pitch.pitchClass, 'name':NoteCollector.terminal_object.name}
        
        initial_object = pd.DataFrame(data=initial_dict, index=[0]) 
        terminal_object = pd.DataFrame(data=terminal_dict, index=[0])
        return initial_object, terminal_object
        
    def __repr__(self):
        return f"NoteCollector {self.order}"
    
    def __str__(self):
        return f"NoteCollector order={self.order} verbose={self.verbose} name={self.name} format={self.format}, source={self.source}"

    def get_base_type(self):
        return note.Note
    
    def set_source(self, source):
        """This method is called first in collect()
        
        Determine if source is a file or folder and if it exists (or not)
        Source can be a single file, or a composer name (such as 'bach')
        A single file must be compressed musicXML (.mxl), uncompressed (.musicxml), or .xml
        An xml file is treated as an uncompressed musicXML file
        A single file that resides in the music21 corpus is specified without an extension.
        The default corpus location ("/Compile/music21/music21/corpus") may be changed by setting corpus_folder attribute
        It may be accessed symbolically on the command line by the $CORPUS variable.
        Examples:
        --source composer='bach'                    # corpus search for composer 'bach'
        --source title='bwv*'                       # corpus wild card search
        --source "composer='bach',title='bwv*'"     # corpus wild card search for composer and title
        --source $CORPUS/haydn/opus74no1            # an individual corpus file
        --source '/data/music/Corpus/dwbzen/Prelude.mxl'    # a single musicXML file. The .mxl may be omitted
        --source [filename].json                    # load a serialized notes DataFrame (TODO)
        
        
        """
        result = True
        title = None
        composer = None
        range_instruments = None
        if self.enforce_range:
            range_instruments = self.instruments
               
        if 'composer' in source or 'title' in source:
            #
            # search the corpus for multiple scores
            # for example: "composer=bach,title=^bwv4+"
            #
            search_string = source.split(",", maxsplit=1)    # title can be a regular expression with embedded commas
            for ss in search_string:
                st = ss.split('=')
                if st[0] == 'composer':
                    composer = st[1]
                elif st[0] == 'title':
                    title = st[1]

            self.scores, self.titles = MusicUtils.get_scores_from_corpus(composer=composer, title=title, keypart=self.key_partName, filters=self.score_filters)
            self.number_of_scores = len(self.scores)
                
            for ascore in self.scores:
                if self.verbose > 0:
                    print(f"working on {ascore.metadata.title}")
                if self.diatonic:
                    #
                    # transpose the score if collection mode is diatonic
                    # default target key of C-Major  or A-minor is used
                    # based on the mode (major or minor) of the original Keys of each part
                    #                    
                    transposed_score = MusicUtils.transpose_score(ascore, partnames=self.part_names, instruments=range_instruments)
                    self.transposed_scores.append(transposed_score)
                    notesdf, pnames, pnums = MusicUtils.get_notes_for_score(transposed_score, self.part_names, self.part_numbers)
                else:
                    #
                    # absolute pitch/pitch class or scale degree.  no transposition, but still need to adjust_accidentals
                    #
                    transposed_score = MusicUtils.adjust_score_accidentals(ascore, partnames=self.part_names, inPlace=False)
                    self.transposed_scores.append(transposed_score)
                    notesdf, pnames, pnums = MusicUtils.get_notes_for_score(transposed_score, self.part_names, self.part_numbers)
                
                if self.verbose > 2 and len(notesdf)>0 and 3 in notesdf['pitchClass']:  # temporary debugging code
                    print(notesdf[notesdf['pitchClass']==3][['part_name','pitch']])
                    
                self.notes_df = pd.concat([self.notes_df, notesdf])  # deprecated:   self.notes_df.append(notesdf)
                if len(pnames) > 0:
                    self.score_partNames = self.score_partNames.union(pnames)
                if len(pnums) > 0:
                    self.score_partNumbers = self.score_partNumbers.union(pnums)
                
            if self.notes_df is None or len(self.notes_df) == 0:
                result = False
        else:
            #
            # a single filename or path - one score
            #
            if source.startswith('$CORPUS'):
                corpus_file = self.corpus_folder + source[6:]
                self.score = corpus.parse(corpus_file)
            else:
                file_info = MusicUtils.get_file_info(source)
                if file_info['Path'].exists():
                    self.score = converter.parse(file_info['path_text'])
                    if self.verbose > 2:
                        self.score.show()
            if self.score is not None:
                self.scores.append(self.score)
                self.number_of_scores = 1
                if self.diatonic:
                    # diatonic pitch - transpose the score to C-Major or C-Minor
                    # and enforce instrument ranges if needed (if enforce_range is True)
                    #
                    self.transposed_score = MusicUtils.transpose_score(self.score, partnames=self.part_names, instruments=range_instruments)
                    self.transposed_scores.append(self.transposed_score)
                    self.notes_df, self.score_partNames, self.score_partNumbers = \
                        MusicUtils.get_notes_for_score(self.transposed_score, self.part_names, self.part_numbers)
                else:   # absolute pitch/pitch class or scale degree - no transposition required
                    self.notes_df, self.score_partNames, self.score_partNumbers = \
                        MusicUtils.get_notes_for_score(self.score, self.part_names, self.part_numbers)
            else:
                result = False

        return result

    def process(self, key_notes:pd.DataFrame, next_note:pd.Series) -> (pd.DataFrame, pd.Series):
        """Adds key_notes and next_note to counts_df DataFrame
            In pitch mode the Note attribute counted is 'nameWithOctave' as in "C4"
            In pitch class mode the attribute is 'name' - so no octave, just the pitch, as in "C"
        
            Args:
                key_notes - a n-row DataFrame from notes_df where n = order of the MarkovChain
                next_note - a single row Series from notes_df that represents the Note following the key_notes
        """
        
        if self.pitch_mode:
            index_str = MusicUtils.show_notes(key_notes,'nameWithOctave')
            col_str = str(next_note.loc['nameWithOctave'])
        elif self.pitch_class_mode:
            index_str = MusicUtils.show_notes(key_notes,'name')
            col_str = str(next_note.loc['name'])
        else:  # self.scale_degree_mode:
            index_str = MusicUtils.show_notes(key_notes,'scaleDegree')
            col_str = str(next_note.loc['scaleDegree'])

        if self.verbose > 1:
            print(f"key_note: {index_str}, next_note: {col_str}")
        
        if self.counts_df is None:
            # initialize the counts DataFrame
            if self.pitch_mode:
                self.counts_df = pd.DataFrame(data=[1],index=[index_str], columns=[next_note.nameWithOctave])
            elif self.pitch_class_mode:
                self.counts_df = pd.DataFrame(data=[1],index=[index_str], columns=[next_note.loc['name']])
            else:  # self.scale_degree_mode:
                self.counts_df = pd.DataFrame(data=[1],index=[index_str], columns=[next_note.loc['scaleDegree']])
        
        else:
            if index_str not in self.counts_df.index:   # add a new row
                self.counts_df.loc[index_str, col_str] = 1
                # self.counts_df.loc[index_str, col_name] = 1
            else: # update existing index
                if col_str in self.counts_df.columns:
                    self.counts_df.loc[index_str, col_str] = 1 + self.counts_df.loc[index_str, col_str]
                else:
                    self.counts_df.loc[index_str, col_str] = 1
        self.counts_df = self.counts_df.fillna(0)
            
    def collect(self) -> MarkovChain:
        """Run collection on the notes_df DataFrame created from the source score(s)
        
        The collection unit is a score part name. Score part_names are in self.score_partNames set.
        The first MarkovChain key entry will consists of the initial_object + the first order-1 notes.
        The last entry will have the terminal_object as the last note.
        Returns MarkovChain result
        """
        if self.source_path is not None:
            self.source = self.set_source(self.source_path)
            
        filename = "{}/{}{}.csv".format(self.save_folder, self.name, self.notes_df_fileName)
        self.notes_df.to_csv(path_or_buf=filename)
        print(f"notes: {filename}")
        
        for pname in self.score_partNames:

            partNotes_df = self.notes_df[self.notes_df['part_name']==pname]
            df_len = len(partNotes_df)
            next_note = None
            key_notes = None
            iloc = 0
            while iloc + self.order < df_len:
                if iloc == 0:
                    key_notes = pd.concat([self.initial_object, partNotes_df[iloc:iloc+self.order-1]] )   # deprecated: self.initial_object.append( partNotes_df[iloc:iloc+self.order-1])
                else:
                    key_notes = partNotes_df.iloc[iloc:iloc+self.order]
                
                next_note = partNotes_df.iloc[iloc+self.order]
                self.process(key_notes, next_note)      # add to counts DataFrame
                
                iloc = iloc + 1
                
            # add the terminal_object to signify end of this part
            key_notes = partNotes_df.iloc[iloc:]
            next_note = self.terminal_object.iloc[0]
            self.process(key_notes, next_note)

        self.counts_df.rename_axis('KEY', inplace=True)
        if self.sort_chain:
            self.counts_df.sort_index('index', ascending=True, inplace=True)
            self.counts_df.sort_index(axis=1, ascending=True, inplace=True)
        #
        # create the MarkovChain from the counts by added probabilities
        #
        sums = self.counts_df.sum(axis=1)
        self.chain_df = self.counts_df.div(sums, axis=0)
        self.chain_df.rename_axis('KEY', inplace=True)
        self.chain_df = self.chain_df.applymap(lambda x: MusicUtils.round_values(x, 6))
        
        self.markovChain = MarkovChain(self.order, self.counts_df,  chain_df=self.chain_df, myname=self.name)
        
        if self.verbose > 1:
            print(f" Counts:\n {self.counts_df}")
            print(f" MarkovChain:\n {self.markovChain}")
            if self.verbose > 1:
                print(self.markovChain.__repr__())
        #
        # collect durations from the Score and notes_df DataFrame
        #
        self.collect_durations(self.notes_df)
        return self.markovChain

    
    def save(self):
        """Saves the chain_df, counts_df, and notes_df DataFrames to a file in specified format.
        
        """
        save_result = super().save()
        if self.notes_df is not None and self.name is not None:
            #
            # optionally save notes_df as .csv
            #
            filename = "{}/{}_notes.{}".format(self.save_folder, self.name, self.format)
            df = self.notes_df[['name','nameWithOctave','pitchClass','ps','part_name','part_number','quarterLength', 'scaleDegree']]
            save_result = CollectorProducer.save_df(df, filename, self.format)
            #df.to_csv(filename)
        return save_result
    
if __name__ == '__main__':
    print(NoteCollector.__doc__)




from common.ruleSet import RuleSet

class SubstitutionSystem(object):
    '''Base substitution system class
    '''
    def __init__(self, rule_set:RuleSet, tags:[str]=None, verbose=0):
        '''Initialize code
            Arguments:
                rule_set - a RuleSet instance
                tags - a list of tags that can appear in substitution rules, default is None
        '''
        self.rule_set = rule_set
        self.substitutions = rule_set.substitutions
        self.tags = tags
        self.verbose = verbose



# ------------------------------------------------------------------------------
# Name:          producer.py
# Purpose:       WordProducer class.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright 2022 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from common.producer import Producer
import random

class WordProducer(Producer):
    
    def __init__(self, state_size, markovChain, source_file, min_size=0, max_size=0, num=10, verbose=0, rand_seed=42):
        super().__init__(state_size, markovChain, source_file, min_size, max_size, num, verbose, rand_seed)
        self.display_as_produced = True
        self.list=False
        self.postprocessing = None
        self.initial=False
        self.pos=False

        self.terminal_object = '~'
        self.initial_object = ' '
        self.output_format = 'TC'   # title case, can also do UC = upper case, LC = lower case
        self._set_keys()
    
    def __str__(self):
        return f"WordProducer order={self.order} verbose={self.verbose}, source={self.source}, min/max={self.min_size},{self.max_size} seed={self.seed}, num={self.num}"
    
    def _set_keys(self):
        """Create the keys
        
        Creates the set of initial keys (_initial_keys) and all keys (_keys)
        from the counts_df DataFrame
        """
        prefix = self.initial_object
        self._initial_keys_df = self.counts_df[ [x.startswith(prefix) for x in self.counts_df['key']] ]
        self._initial_keys = list(set(self._initial_keys_df['key'].values.tolist()))
        
        self._keys = list(set(self.counts_df['key'].values.tolist()))

        if self.verbose > 1:
            print(f'total number of keys: {len(self._keys)}')
            print(f'number of initial keys: {len(self._initial_keys)}')
            
    def get_seed(self, aseed=None):
        if self.seed is None or aseed is None:
            # need to pick one at random
            if self.initial:
                # pick a seed that starts a word (i.e. first character is a blank)
                theseed = self._initial_keys[random.randint(0, len(self._initial_keys)-1)]
            else:
                # pick any old seed from the chain's index
                theseed = self._keys[random.randint(0, len(self._keys)-1)]
        else:
            theseed = aseed
        return theseed

    
    def get_next_character(self, seed):
        """Gets the next character based on the seed argument
        
        Returns:
            both the next character and a new seed as a dict with keys 'new_seed' and 'next_token'
        """
        
        new_seed = None
        next_token = None
        if self.verbose > 2:
            print(f"get_next_character(seed): '{seed}'")
            
        if seed in self._keys:
            df = self.chain_df[self.chain_df['key']==seed]
            if len(df) == 0:  #  add to chain_df from counts_df
                df = self._add_key(seed)
            
            #row = self.chain_df.loc[seed]
            #row_probs = row[row > 0].cumsum()
            # given a probability, pick the first character in the row_probs (cumulative probabilities)
            # having a probability > the random probability
            # for example:
            prob = random.random()
            row = df[df['prob']>prob].iloc[0]
            next_token = row['word']
            p = row['prob']
            #p = row_probs[row_probs> prob].iat[0]
            #next_token = row_probs[row_probs> prob].index[0]
            new_seed = (seed + next_token)[1:self.order+1]
            if self.verbose > 1:
                print(f"random prob: {prob}, row prob: {p},  seed: '{seed}', next_token: '{next_token}', new_seed: '{new_seed}'")
        else:
            if self.verbose > 0:
                print(f"seed: '{seed}' not in keys")
                
        return dict([ ('next_token', next_token), ('new_seed',new_seed)])
    
    def add_word_to_list(self, word_list, aword):
        if self.output_format == 'TC':
            aword = aword.title()
        elif self.output_format == 'LC':
            aword = aword.lower()
        elif self.output_format == 'UC':
            aword = aword.upper()
        else:
            pass
        word_list.append(aword)
        if self.display_as_produced:
            print(f"{aword}")
        return aword

    def produce(self):
        """
        Produce a list of words using the MarkovChain probabilities
        """
        words = []
        initial_seed = self.get_seed(self.seed)      # get the initial seed
        seed = initial_seed
        self.seed = seed
        if self.verbose > 0:
            print(f"initial seed: '{seed}'")
        if self.verbose > 2:
            print(f" MarkovChain:\n {self.markovChain}")
        for n in range(self.num):
            word = seed.strip()
            more_to_go = True
            if self.verbose > 1:
                print(f"generating word {n}")
            while more_to_go:
                nc_dict = self.get_next_character(seed)
                ntoken = nc_dict['next_token']
                nseed = nc_dict['new_seed']
                if ntoken is not None:
                    if ntoken==self.terminal_object:     # done with this word
                        if len(word) < self.min_size:
                            pass
                        else:
                            word = self.add_word_to_list(words, word)
                            more_to_go = False
                            seed = self.get_next_seed()
                    else:
                        word = word + ntoken
                        if len(word) >= self.max_size:
                            word = self.add_word_to_list(words, word)
                            more_to_go = False
                            seed = self.get_next_seed()
                        else:
                            more_to_go = True
                            seed = nseed
                else:
                    # there is no next token for the current seed, so done with this word
                    word = self.add_word_to_list(words, word)
                    more_to_go = False
                    seed = initial_seed             # start next word with the initial_seed selected earlier
            
        if self.sort:
            words = words.sort(key=str.lower, reverse=False)
        return words

if __name__ == '__main__':
    print(WordProducer.__doc__)




# ------------------------------------------------------------------------------
# Name:          producer.py
# Purpose:       Runs CharacterCollector if needed and then WordProducer
#
#
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright 2021 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from common.characterCollector import CharacterCollector
from common.utils import Utils
from common.markovChain import MarkovChain
from common.wordProducer import WordProducer
import pandas as pd
import argparse

class WordProducerRunner(object):
    
    if __name__ == '__main__':
        parser = argparse.ArgumentParser()
        #
        # CharacterCollector arguments
        #
        parser.add_argument("order", help="the order of the Markov Chain", type=int, choices=range(1,5))
        parser.add_argument("-t", "--text", help="in-line text input. One of --text or --source must be specified")
        parser.add_argument("-s", "--source", help="input file name")
        parser.add_argument("-i","--ignoreCase", help="ignore input case", action="store_true", default=False)
        parser.add_argument("--name", help="Name of resulting MarkovChain, used to save to file", type=str, default="mychain")

        #
        # WordProducer arguments
        #
        parser.add_argument("-c", "--chainfile", help="Existing serialized MarkovChain file, or POS file name.",  type=str, default=None)
        parser.add_argument("-f","--format", help="File input/output format. Default is csv", type=str, choices=['csv','json','xlsx'], default='csv' )

        parser.add_argument("-n", "--num", help="Number of words to produce",  type=int, default=10)
        parser.add_argument("--min", help="Minimum length of words", type=int, choices=range(2,6), default=4)
        parser.add_argument("--max", help="Maximum length of words", type=int, choices=range(3,20), default=10)
        parser.add_argument("-v","--verbose", help="increase output verbosity", action="count", default=0)
        parser.add_argument("--seed", help="Initial seed, length must be equal to the order of source chain", default=None)
        parser.add_argument("--sort", help="Sort the output in Ascending (A) or Decending (D) order", default=None)
        parser.add_argument("-l", "--list", help="Display produce Words in order produced, default is false", action="store_true",default=False)
        parser.add_argument("-p", "--postprocessing", help="post-processing: TC = title case, UC = upper case, LC = lower case. Default is None", default=None)
        parser.add_argument("--initial", help="choose initial seed only (start of word)", action="store_true", default=False)
        parser.add_argument("--pos", help="Source is a part-of-speech file", action="store_true",default=False)
        parser.add_argument("--recycle", help="How often to pick a new seed, default is pick a new seed after each word produced", type=int, default=1)
        parser.add_argument("--display", "-d", help="Display each word as it is produced", action="store_true", default=True)
        args = parser.parse_args()
        
        markovChain = None
        order = args.order
        source_file = None
        chain_filename = None
        counts_filename = None
        display_as_produced = args.display
        if args.verbose > 0:
            print('run WordProducer')
            print(args)
        if args.chainfile is None and not (args.source is None and args.text is None):   # run the collector first
            source_file = args.source
            collector = CharacterCollector(state_size = args.order, verbose=args.verbose, source=source_file, text=args.text, ignore_case=args.ignoreCase)
            collector.sort_chain = True
            run_results = collector.run()
            if run_results['save_result']:
                print(f"MarkovChain written to file: {collector.filename}")
            
            markovChain = collector.markovChain
            
        else:
            # use serialized MarkovChain file  and counts file in specified format
            # for example  --chainfile "/Compile/dwbzen/resources/text/drugBrands" 
            #
            order_string = '_0{}'.format(order)
            chain_filename = f'{args.chainfile}_charsChain{order_string}.{args.format}'
            counts_filename = f'{args.chainfile}_charCounts{order_string}.{args.format}'
            
            chain_file_info = Utils.get_file_info(chain_filename)
            chain_path = chain_file_info["path_text"]
            
            counts_file_info = Utils.get_file_info(counts_filename)
            counts_path = counts_file_info["path_text"]

            ext = args.format
            if args.verbose > 0:
                print(chain_file_info)
            if not chain_file_info['exists']:
                print(f"{chain_path} does not exist")
                exit()
            else:
                if ext=='json':
                    chain_df = pd.read_json(chain_path, orient="index")
                    counts_df = pd.read_json(counts_path, orient="index")
                    markovChain = MarkovChain(order, counts_df, chain_df=chain_df)
                elif ext=='csv':     # parts-of-speech file  TODO
                    chain_df = pd.read_csv(chain_path, header=0, names=['key','word','count','prob'])
                    counts_df = pd.read_csv(counts_path, header=0, names=['key','word','count'])
                    markovChain = MarkovChain(order, counts_df, chain_df=chain_df)
                elif ext=='xlsx':
                    pass        # TODO

        #
        # markovChain will never, ever be None
        #
        wordProducer = WordProducer(order, markovChain, source_file, args.min, args.max, args.num, args.verbose )
        if chain_filename is not None:
            wordProducer.chain_filename = chain_filename
            wordProducer.counts_filename = counts_filename

        wordProducer.initial = args.initial
        wordProducer.seed = args.seed
        wordProducer.recycle_seed_count = args.recycle
        wordProducer.display_as_produced = display_as_produced
        words = wordProducer.produce()  # also prints the words
        if not display_as_produced:
            print(words)
            




# ------------------------------------------------------------------------------
# Name:          intervalCollectorRunner.py
# Purpose:       Run IntervalCollector
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2021 Donald Bacon

# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from common.collector import Collector
from music.intervalCollector import IntervalCollector
import argparse

class IntervalCollectorRunner(Collector):
    
    if __name__ == '__main__':
        parser = argparse.ArgumentParser()
        parser.add_argument("--order", "-o", help="the order of the Markov Chain state space", type=int, choices=range(1,5))
        parser.add_argument("-s", "--source", help="input file (.mxl or .musicxml), folder name, or composer name ('bach' for example)")
        parser.add_argument("-v","--verbose", help="increase output verbosity", action="count", default=0)
        parser.add_argument("-n","--name", help="Name of resulting MarkovChain, used to save to file", type=str)
        parser.add_argument("-f","--format", help="Save output format. Default is csv", type=str, choices=['csv','json','xlsx'], default='json' )
        parser.add_argument("--sort", help="Sort resulting MarkovChain ascending on both axes", action="store_true", default=False)
        parser.add_argument("-d","--display", help="display resulting MarkovChain in json or cvs format", type=str, choices=['csv','json','chain'] )
        parser.add_argument("-p","--parts", help="part name(s) or number(s) to include in building the MarkovChain", type=str)
        parser.add_argument("--filter", help="Apply filter to parts", type=str, default=None)
        
        args = parser.parse_args()
        if args.verbose > 0:
            print('run IntervalCollector')
            print(args)

        collector = IntervalCollector(state_size = args.order, verbose=args.verbose, source=args.source, parts=args.parts )
        collector.name = args.name
        collector.format = args.format
        collector.sort_chain = args.sort
        collector.set_score_filter(args.filter)    # could be None
        #
        # only for Bach works in the music21 corpus
        #
        collector.key_partName = 'Soprano'
                
        if args.verbose > 0:
            print(collector.__str__())

        
        run_results = collector.run()
        if run_results['save_result']:
            print(f"MarkovChain written to file: {collector.filename}")
        if args.display is not None:
            # display in specified format
            if args.display=='json':
                print(collector.get_json_output())
            elif args.display=='csv':
                print(collector.markovChain.chain_df.to_csv(line_terminator='\n'))
            else:   # assume 'chain' and display the data frame
                print(collector.markovChain.chain_df)




from common.ruleSet import RuleSet
from music.scoreGen import ScoreGen
from music.musicSubstitutionRules import MusicSubstitutionRules
from music.musicSubstitutionSystem import MusicSubstitutionSystem
from music.musicUtils import MusicUtils
from music21 import key, note
import argparse, json
from common.environment import Environment

class ScoreGenRunner(object):
    '''Runs MusicSustitutionSystem and passes the result to ScoreGen

    '''
    if __name__ == '__main__':
        
        env = Environment.get_environment()
        resource_folder = env.get_resource_folder('music')
        start = ['0/1.0', '+1/0.5', '-2/1.0', '1/2.0']
        parser = argparse.ArgumentParser()
        parser.add_argument("--scale", "-s", help="Scale name", type=str, default='Major')
        parser.add_argument("--parts", "-p", help='Part name(s)', type=str,  nargs='*', default=['Soprano'])
        parser.add_argument("--key", "-k", help="Key, default is C-major", type=str, default='C')
        parser.add_argument("--steps", "-n", help="number of steps to run", type=int, choices=range(0,10), default=1)
        parser.add_argument("--start", help="starting commands", type=str, nargs='*', default=start)
        parser.add_argument("--notes", help="Starting note(s) with octave", type=str, nargs='*', default=["C5"])
        parser.add_argument("--duration", "-d", help="Starting note duration in quarterLengths", type=float, default=4.0)
        parser.add_argument("--show", help="How to view the resulting score, default is musicxml", type=str, default='musicxml')
        parser.add_argument("-v","--verbose", help="increase output verbosity", action="count", default=0)
        parser.add_argument("--trace", "-t", help="Trace rule processing", action="store_true", default=False)
        #
        # specifies substitution rules JSON file and the substitution rule name
        # default file is 'resources/music/substitution_rules.json'
        # Need to specify a rule name for each part OR the hard-coded substitution_rules is used for all parts
        #
        parser.add_argument("--rules", "-r", help="Substitution rules JSON file name", type=str, default=f'{resource_folder}/substitution_rules.json')
        parser.add_argument("--rulenames", help="Rule name(s) in Substitution rules JSON file.", type=str, nargs='*', default=None)
        args = parser.parse_args()
        
        #
        # default hard-coded rule set to use if Substitution rules name is not specified
        #
        substitution_rules = {\
            r'(?P<interval>[+-]?[01])/(?P<duration>\d+\.\d+)':['0/0.5', '+3/1.0', '-2/2.0', '-1/2.0']  , \
            r'(?P<interval>[+-]?[23])/(?P<duration>\d+\.\d+)' : ['-1/1.0', '-2/2.0', '+2/0.5', '0/1.0']
        }

        #
        # generate for each part
        #
        ascore = None
        for partnum in range(len(args.parts)):
            partname = args.parts[partnum]
            if args.rulenames is not None:
                fp = open(args.rules, "r")
                jtxt = fp.read()
                jdoc = json.loads(jtxt)
                substitution_rules = jdoc[args.rulenames[partnum]]
                
            command_rules = {'interval' : MusicSubstitutionRules.interval_rule, 'duration' : MusicSubstitutionRules.duration_rule}
            rules = {'commands':command_rules}
            splitter = r'(?P<interval>[+-]?\d+)/(?P<duration>\d+\.\d+)'
            rule_set = RuleSet(substitution_rules, rules=rules, splitter=splitter)
            
            MusicSubstitutionRules.trace = args.trace
            MusicUtils.verbose = args.verbose
            ss = MusicSubstitutionSystem(rule_set, verbose=args.verbose)
            commands = ss.apply(start,args.steps)
            print(f'{len(commands)} commands:\n{commands} \n')
            
            score_gen = ScoreGen(rule_set, scale_name=args.scale, instrument_name=partname, key=key.Key(args.key), verbose=args.verbose)
            start_note=note.Note(args.notes[partnum], quarterLength=args.duration)
            score_gen.run(commands, start_note=start_note)
            if ascore is None:
                maxpartlen = score_gen.part.duration
                ascore = score_gen.score
            else:
                if maxpartlen.quarterLength < score_gen.part.duration.quarterLength:
                    maxpartlen = score_gen.part.duration
                ascore.append(score_gen.part)
                
        MusicUtils.extend_parts(ascore, maxpartlen)
        ascore.show(args.show)
        
        


# ------------------------------------------------------------------------------
# Name:          scale.py
# Purpose:       Encapsulate a musical scale.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2022 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------
import json
    
def find_note_in_scale(note):
    # returns the index of note in one of the CHROMATIC scales or -1 if not there
    _note = note.upper()
    _index = -1
    if(Scale.has_octave(note)):
        _note = _note[0:len(note)-1]
    if _note in Scale.CHROMATIC_SHARP_PITCHES:
        _index = Scale.CHROMATIC_SHARP_PITCHES.index(_note)
    elif _note in Scale.CHROMATIC_FLAT_PITCHES:
        _index = Scale.CHROMATIC_SHARP_PITCHES.index(_note)
    return _index

class Scale:
    """
    A single Scale as a dictionary with keys:
    name - primary lookup key
    alternateNames - list of AKAs
    groups - list of group names the scale is a member of
    formula - a list of integer intervals that define the scale
    size - number of notes in the scale
    """
    
    CHROMATIC_SHARP_PITCHES = ['C','C#','D','D#','E','F','F#','G','G#','A','A#','B']
    CHROMATIC_FLAT_PITCHES =  ['C','Db','D','Eb','E','F','Gb','G','Ab','A','Bb','B']
    CHROMATIC_MIXED_PITCHES = ['C','C#','D','Eb','E','F','F#','G','G#','A','Bb','B']
    

    @classmethod
    def has_octave(cls, note):
        _result = len(note) > 0 and note[len(note)-1].isdigit()
        return _result

    @classmethod
    def get_octave(cls, note):
        if(Scale.has_octave(note)):
            return  note[len(note)-1]
        else:
            return None

    @classmethod
    def get_pitch(cls, note):
        _pitch = note
        if(Scale.has_octave(note)):
            _pitch = _pitch[0:len(note)-1]
        return _pitch
        
    def __init__(self, raw_data={}):
        '''
        Constructor
        '''
        # print("raw data: {}".format(raw_data))
        if(type(raw_data) is dict):
            self.scale = raw_data
        else:
            self.scale = json.loads(raw_data)
            
        self.name = self.scale['name']
        self.formula = self.scale['formula']
        self.size = len(self.formula)
        Scale.chromatic_sharp_pitches2 = Scale.CHROMATIC_SHARP_PITCHES.copy()
        Scale.chromatic_flat_pitches2 = Scale.CHROMATIC_FLAT_PITCHES.copy()
        Scale.chromatic_mixed_pitches2 = Scale.CHROMATIC_MIXED_PITCHES.copy()
        Scale.chromatic_sharp_pitches2 *=2  # two octaves
        Scale.chromatic_flat_pitches2 *=2    # two octaves
        Scale.chromatic_mixed_pitches2 *=2   # two octaves
        self._chromatic_scale = []

    def __iter__(self):
        """ Iterate over the notes in the scale """
        return self.scale.__iter__()
        

    def pitches(self, root='C4', accidental_preference=None):
        """
        Creates and returns the list of notes for this scale starting with the root provided. the default root is 'C4' (middle C).
        If the root provided lacks an octave, 4 is assumed.
        By convention, pitches are expressed in scientific music notation which includes an octave designation starting at 0
        For example, the range of a Piano is A0 to C8.
        Octaves are from C to B, so the pitch following B4 is C5.
        In general the accidental (# or b) used is determined first by the root note. If the root doesn't have an accidental,
        the accidental_preference is used. If a preference is not provided a # is the default.
        Preferences: 'sharp', '#', 'flat', 'b', 'mixed', '#b'
        """
        _ap = self.get_accidental_preference(root, accidental_preference)
        _p = Scale.get_pitch(root)
        _notes = self.notes(_p, accidental_preference)
        if not Scale.has_octave(root):
            return _notes
        #
        # add the octave to each note
        #
        _octave = int(Scale.get_octave(root))
        _ind = self._chromatic_scale.index(_p)
        _pitches = []
        i = 0
        next_octave = True
        for note in _notes:
            if(_ind > 11 and next_octave):
                _octave = _octave + 1
                next_octave = False
            _pitches.append('{}{}'.format(note,_octave))
            if(i < len(self.formula)):
                _ind = _ind + self.formula[i]
            i = i + 1
        return _pitches


    def notes(self, root = 'C', accidental_preference=None):
        """
        Creates and returns the list of notes for this scale starting with the root note provided. the default root is 'C'. 
        By convention, notes do not include an octave designation.
        """
        _root = root
        if(Scale.has_octave(root)):
            _root = root[0:len(root)-2]
        _notes = [_root]
        _ap = self.get_accidental_preference(root, accidental_preference)
        _ind = self._chromatic_scale.index(_root)
        for i in self.formula:
            _ind += i
            _notes.append(self._chromatic_scale[_ind])
        return _notes
    
    def get_accidental_preference(self, root, accidental_preference):
        _ap = '#'
        if accidental_preference is None:
            if('#' in root):
                _ap = '#'
                self._chromatic_scale=Scale.chromatic_sharp_pitches2
            elif('b' in root):
                _ap = 'b'
                self._chromatic_scale=Scale.chromatic_flat_pitches2
            else:
                _ap = '#b'
                self._chromatic_scale=Scale.chromatic_mixed_pitches2
        else:
            if('#' == accidental_preference or 'sharp' == accidental_preference):
                _ap = '#'
                self._chromatic_scale=Scale.chromatic_sharp_pitches2
            elif('b' == accidental_preference or 'flat' == accidental_preference):
                _ap = 'b'
                self._chromatic_scale=Scale.chromatic_flat_pitches2
            elif('#b' == accidental_preference or 'mixed' == accidental_preference):
                _ap = '#b'
                self._chromatic_scale=Scale.chromatic_mixed_pitches2
            else:
                _ap = '#b'
                self._chromatic_scale=Scale.chromatic_mixed_pitches2
        return _ap
    
    def __str__(self):
        return str(self.notes(root='A', accidental_preference='mixed'))
    
    def __repr__(self):
        return str(self.scale)

if __name__ == '__main__':
    print(Scale.__doc__)

        
        
        


'''
Created on Mar 2, 2021

@author: don_bacon
'''

import json
from music.chord import Chord
import music
from pandas.core.frame import DataFrame

class Chords(object):
    '''
    Chord formulas
    '''

    def __init__(self, raw_data={}, json_file_name=''):
        '''
        Constructor
        '''
        self.chords = {}
        self.chord = {}
        self.chords_file = ''
        self.data = {}
        self.chord_names = []
        if (len(json_file_name) >0):
            self.chords_file = json_file_name
            with open(self.chords_file, "r") as read_file:
                self.data = json.load(read_file)
        else:
            self.data = raw_data
            
        self.df_chords = self._get_chords_data()
    
    def _get_chords_data(self):
        self.chords = self.data['chordFormulas']
        for s in self.chords:
            self.chord_names.append(s['name'])
            self.chord[s['name']] = Chord(s)
        return DataFrame(self.chords)
    
    def get_chords(self):
        return self.df_chords
    
    def __iter__(self):
        ''' Iterate over the chords '''
        return self.chords.__iter__()
    
    if __name__ == '__main__':
        print('Sample Chords usage')
        chords_file = "/Compile/dwbzen/resources/music/allChordFormulas.json"
        _chords = music.Chords(json_file_name = chords_file)
        _chord = _chords.chord['Ninth flat fifth']
        print(_chord.chord)
        print('Ninth flat fifth: {}'.format(_chord.notes))
        
        # get the notes for a different root
        _notes = _chord.scale.notes(root='A', accidental_preference='mixed')
        print(_notes)
        
        

    


from __future__ import absolute_import  # multi-line and relative/absolute imports

from .__version__ import __title__, __description__, __version__
from .__version__ import __author__, __author_email__, __license__
from .__version__ import __copyright__


__all__ = [
    'chord', 
    'chords', 
    'durationCollector', 
    'instruments', 
    'intervalCollector',
    'intervalCollectorRunner',
    'keys',
    'musicCollector',
    'musicProducer',
    'musicProducerRunner',
    'musicScale',
    'musicSubstitutionRules',
    'musicSubstitutionSystem',
    'musicUtils',
    'noteCollector',
    'noteCollectorRunner',
    'sample_usage',
    'scale',
    'scales',
    'scoreGen',
    'scoreGenRunner',
    'song'
]

from dwbzen.music.chords import Chords
from dwbzen.music.chord import Chord
from dwbzen.music.durationCollector import DurationCollector
from dwbzen.music.instruments import Instruments
from dwbzen.music.intervalCollector import IntervalCollector
from dwbzen.music.intervalCollectorRunner import IntervalCollectorRunner
from dwbzen.music.keys import Keys
from dwbzen.music.musicCollector import MusicCollector
from dwbzen.music.musicProducer import MusicProducer
from dwbzen.music.musicProducerRunner import MusicProducerRunner
from dwbzen.music.musicScale import MusicScale
from dwbzen.music.musicSubstitutionSystem import MusicSubstitutionSystem
from dwbzen.music.musicSubstitutionRules import MusicSubstitutionRules
from dwbzen.music.musicUtils import MusicUtils
from dwbzen.music.noteCollector import NoteCollector
from dwbzen.music.noteCollectorRunner import NoteCollectorRunner
from dwbzen.music.sample_usage import keys_usage, scales_usage, song_usage
from dwbzen.music.scale  import Scale
from dwbzen.music.scales import Scales
from dwbzen.music.scoreGen import ScoreGen
from dwbzen.music.scoreGenRunner import ScoreGenRunner
from dwbzen.music.song import Song





import re

class RuleSet(object):

    def __init__(self, substitution_rules:dict, rules=None, splitter:str=None):
        '''
        Each substitution is a dict. The key is a regular expression string used to match an input string,
        for example: (interval)/(quarterLength-multiplier)
        The value is the replacement. The re groups are named and can be referenced in the replacement.
        splitter is a regular expression, analagous to a string splitter, that will split the
        substitution result into named components.
        
        Sample substitutions:
           sub1 = {r'(?P<interval>[+-]?0)/(?P<duration>\d+\.\d+)':['0/0.5', '+1/1.0', '-2/1.0', '+1/2.0']}
           sub2 = {r'(?P<interval>[+-]?1)/(?P<duration>\d+\.\d+)' : ['0/<duration>', '1/0.5']}
           splitter = r'(?P<interval>[+-]?\d+)/(?P<duration>\d+\.\d+)'
        '''
        self.rules = rules
        self.substitutions = []
        self.substitution_rules = substitution_rules

        #
        # compile the substitution rules regular expressions
        #
        for k in self.substitution_rules.keys():
            pattern = re.compile(k)
            replacement = self.substitution_rules[k]
            self.substitutions.append({'pattern':pattern, 'replacement':replacement})
        
        self.pre_processing = None
        self.post_processing = None
        self.command_rules = None
        if splitter is not None:
            self.splitter = re.compile(splitter)
        if rules is not None:
                if 'preProcessing' in rules:
                    self.pre_processing = rules['preProcessing'] 
                if 'postProcessing' in rules:
                    self.post_processing = rules['postProcessing']
                if 'commands' in rules:
                    self.command_rules = rules['commands']
    


# ------------------------------------------------------------------------------
# Name:          producer.py
# Purpose:       Base Producer class.
#
#
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright 2022 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

import random
from common.markovChain import MarkovChain
from common.collectorProducer import CollectorProducer
import pandas as pd

class Producer(CollectorProducer):
    
    def __init__(self, state_size, markovChain:MarkovChain, source=None, min_size=0, max_size=0, num=20, verbose=0, domain='text', rand_seed=None):
        
        super().__init__(state_size, verbose=verbose, source=source, domain=domain)
        
        self.markovChain = markovChain        # a MarkovChain instance, provided by a ProducerRunner
        self.min_size = min_size
        self.max_size = max_size
        self.num = num      # units defined in Producer subclass
        
        self.outfile=None
        self.seed=None
        self.sort=None
        self.initial=True
        self.chain_df = self.markovChain.chain_df
        self.counts_df = self.markovChain.counts_df
        self.chain_updated = False
        
        #
        # initial keys and all keys
        #
        self._initial_keys = None
        self._keys = None
        self._initial_keys_df = None
        self.display_as_produced = False   # display each object (word, sentence etc.)  as it's produced
        #
        # count of how many times the current seed has been used
        # when >= self.recycle_seed_count, a new seed is picked
        self.seed_count = 0
        self.recycle_seed_count = 1     # pick a new seed every n things produced

        random.seed(a=rand_seed)
        
        
    def __repr__(self):
        return f"Producer {self.order}"
    
    def _add_key(self, key):
        key_df = self.counts_df[self.counts_df['key']==key]
        s = (key_df['count']/key_df['count'].sum()).cumsum()
        key_df = key_df.assign(prob=s )
        self.chain_df = pd.concat([self.chain_df, key_df], ignore_index= True)
        self.chain_updated = True   # this will trigger a save
        return key_df
    
    def get_seed(self):  # override in derived class
        return None
    
    def get_next_seed(self):
        """Checks the seed_count against the recycle_seed_count and if >= gets a new seed, otherwise returns the existing one
        
        """
        self.seed_count = self.seed_count + 1
        aseed = self.seed
        if self.seed_count >= self.recycle_seed_count or aseed is None:
            # pick a new seed
            aseed = self.get_seed()
            self.seed_count = 0
            self.seed = aseed
        return aseed
    
    def produce(self):
        """
        Run the Producer with the parameters provided
        Returns a list of whatever the specific Producer is, well, producing (str for example)
        Override this in derived classes
        """
        return None

    def save(self):
        save_result = super().save()
        return save_result
    



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# Name:          markovChain.py
# Purpose:       
#
# Authors:       Donald Bacon
#
# Copyright:    Copyright © 2022 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

'''
 MarkovChain is an implementation of a discreet-time Markov Chain (DTMC) with memory.
 It undergoes transitions from one state to another on a state space, 
 with the probability distribution of the next state depending only on the current state
 and not on the sequence of events that preceded it.
 The states are T instances, transitions from state T1 to state T2 are K instances
 each transition has an associated probability and additional information
 used by classes that generate Ts.</p>
  
 The order of the MarkovChain is the number of states that determine future states.
'''

import pandas as pd

class MarkovChain(object):

    def __init__(self, state_size, counts_df:pd.DataFrame, keys=None, chain_df:pd.DataFrame=None, myname:str=None):
        """Create and initialize a MarkovChain from a DataFrame or dict
        
        Either chain_df or chain_dict must be not None.
        Parameters:
            state_size - the MarkovChain order, also the logical size of the keys
            counts_df - a counts DataFrame, None is okay
            chain_df - if not None, an existing chain DataFrame
            keys - keys to use. If None, keys are derived from counts_df 'key' column.
            myname - optional name to use when saving
        """
        self.order = state_size
        self.counts_df = counts_df
        self.name = myname          # used when persisting to CSV, Excel, or JSON
        self._keys = keys
        self.chain_df = chain_df
        if chain_df is None:
            self._create_chain()        # create the chain_df DataFrame from the counts_df DataFrame
        else:
            self.chain_df = chain_df    # an existing chain_df is provided

    
    def __repr__(self, *args, **kwargs):
        return self.chain_df.to_json(path_or_buf = None,orient='index')
    
    def __str__(self):
        return self.chain_df.to_csv(path_or_buf = None, line_terminator='\n')
    
    def _create_chain(self):
        if self._keys is None:
            self._keys = set(self.counts_df['key'].values.tolist())
        
        for akey in self._keys:
            key_df = self.counts_df[self.counts_df['key'] ==  akey]
            s = (key_df['count']/key_df['count'].sum()).cumsum()
            model = key_df.assign(prob=s )
            if self.chain_df is None:
                self.chain_df = pd.DataFrame(model)
            else:
                self.chain_df = pd.concat([self.chain_df, model], ignore_index= True)
        

    def get_rows(self, key) -> pd.DataFrame:
        """Returns the rows of the chain DataFrame for a given key
        If key not present, return DataFrame will have len==0 
        """
        return self.chain_df[self.chain_df['key'] ==key]
        
if __name__ == '__main__':
        print(MarkovChain.__doc__())





# ------------------------------------------------------------------------------
# Name:          utils.py
# Purpose:       Common utilities.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2022 Donald Bacon

# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

import pathlib
import json
import math
from datetime import datetime

class Utils(object):
    
    known_extensions = ['json', 'txt', 'mxl', 'xml', 'musicxml', 'csv']

    @staticmethod
    def get_file_info(cpath, def_extension='json'):
        """breaks up a path name into component parts
        
        """
        
        x = cpath.split("/")
        paths = x[0:len(x)-1]
        filename = x[-1]
        ext = filename.split(".")       # name.extension as in "foo.csv"
        name = ext[0]
        if len(ext)==2 and ext[1] in Utils.known_extensions:
            ext = ext[1]
            path = cpath      
        else:
            ext = def_extension
            filename = f"{filename}.{ext}"
            path = f"{cpath}.{ext}"
        p = pathlib.Path(path)
        exists = p.exists()
        return  {'paths':paths, 'path_text':path, 'filename':filename, 'name':name,'extension': ext, 'Path':p, 'exists':exists}


    @staticmethod
    def round_values(x, places=5):
        if not type(x) is str:
            return round(x, places)
        else:
            return x
    
    @staticmethod
    def get_json_output(df, orient='index'):
        """Dumps the dataframe argument to a nicely formated text string.
        
        """
        result = df.to_json(orient=orient)
        parsed = json.loads(result)
        return json.dumps(parsed, indent=4)
    
    @staticmethod
    def time_since(base_date:datetime=datetime(2000, 1, 1, 0, 0),  end_date=datetime.now(), what='seconds') -> int:
        """Gets the elapsed seconds or days between two given base date/datetime
            Arguments:
                base_date : the base date, default is 12:00 AM 2000-01-01
                end_date : the ending date/datetime, default is now()
                what : 'seconds' or 'days'
            Returns:
                The seconds or days since the base date to end_date as an integer
            Note - useful in setting a random seed as in random.seed(time_since())
        """
        
        delta = end_date-base_date
        if what=='seconds':
            return math.trunc(delta.total_seconds())
        else:   # assume days
            return delta.days
        
if __name__ == '__main__':
    print(Utils.__doc__)
    



from common.ruleSet import RuleSet
from common.substitutionSystem import SubstitutionSystem
import music
import argparse

class MusicSubstitutionSystem(SubstitutionSystem):
    '''Generates a list representing the evolution of the substitution system with the specified rule 
        from initial condition init for t steps.
        Arguments:
            substitution_rules - 
            start - initial condition
            t - the number of steps (generations) as a range(n1, n2)
                where n1 is the starting step, and n2-1 is the ending step, 0 <= n1 <= n2
                For example, t=range(1, 4) will output the results of steps 1,2 and 3
                t=range(2, 3) step 2 only.
                t=range(2, 4) steps 2 and 3
    '''
    
    def __init__(self, rule_set:RuleSet, verbose=0):
        super().__init__(rule_set, tags=['interval', 'duration'], verbose=verbose)

    
    def apply(self, start:[str], nsteps):
        '''Apply the substitution to the start string for a given number of steps.
         The result is a command list that can be input to a music generation module.
        '''
        sub_result = start
        self.steps = nsteps
        if nsteps > 0:
            sub_result = start
            for step in range(nsteps):
                sub_result = self.apply_step(sub_result)
                if self.verbose > 1:
                    print(f'step: {step}\nresult: {sub_result}')
        return sub_result
            
        
    def apply_step(self, start:[str]):
        subst_result = []
        for subs in self.substitutions:
            pattern = subs["pattern"]
            replacement = subs["replacement"]

            # try to match each item in start (result) list
            # 
            for s in start:
                if self.verbose > 1:
                    print(f's: "{s}"    pattern: {pattern.pattern}')
                match = pattern.match(s)
                if match is not None:
                    if self.verbose > 0:
                        print(f'"{s}" matched pattern {pattern.pattern} ')
                    grp_dict = match.groupdict()
                    #
                    # now substitute
                    #
                    subst_result += replacement
                else:
                    subst_result += [s]
            if self.verbose > 0:
                print(f'step result: {subst_result} \n')
        return subst_result
    
    if __name__ == '__main__':
        
        parser = argparse.ArgumentParser()
        parser.add_argument("--steps", "-s", help="number of steps to run", type=int, choices=range(1,10), default=1)
        parser.add_argument("-v","--verbose", help="increase output verbosity", action="count", default=0)
        args = parser.parse_args()
        
        substitution_rules = {\
            r'(?P<interval>[+-]?0)/(?P<duration>\d+\.\d+)':['0/0.5', '+1/1.0', '-2/1.0', '+1/2.0'], \
            r'(?P<interval>[+-]?1)/(?P<duration>\d+\.\d+)' : ['0/1.0', '1/0.5', '2/1.0']        }
        command_rules = {'interval' : music.MusicSubstitutionRules.interval_rule, 'duration' : music.MusicSubstitutionRules.duration_rule}
        rules = {'commands':command_rules}
        splitter = r'(?P<interval>[+-]?\d+)/(?P<duration>\d+\.\d+)'
        rule_set = RuleSet(substitution_rules, rules=rules, splitter=splitter)
        start = ['0/1.0', '+1/0.5']
        
        ss = music.MusicSubstitutionSystem(rule_set, verbose=args.verbose)
        commands = ss.apply(start,args.steps)
        print(f'{len(commands)} commands:\n{commands} ')
    
    



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# Name:          durationCollector.py
# Purpose:       DurationCollectorr class.
#
#                DurationCollector creates a MarkovChain of durations from a Note stream.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2022 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from music.musicCollector import MusicCollector
from music.musicUtils import MusicUtils
from common.markovChain import MarkovChain
import pandas as pd
import sys
from music21 import duration


class DurationCollector(MusicCollector):

    def __init__(self, state_size=2, verbose=0, source=None, parts=None):          # this also executes self.set_source()
        super().__init__(state_size, verbose, source)
        
        self.terminal_object = duration.Duration(24.0)      # equivalent to 6 x 4 quarter notes
        self.initial_object = duration.Duration(0.0)        # 0 duration have ordinal None
        self.score = None       # if source is a single Score
        self.part_names = []
        self.part_numbers = []
        self.parts = None
        self.durations_df = None
        self.source_df = None
        self.countsFileName = '_durationsCounts' + '_0{}'.format(state_size)
        self.chainFileName = '_durationsChain' + '_0{}'.format(state_size)
        if parts is not None:
            self.add_parts(parts)
        if source is not None:
            self.source = self.set_source(source)   # otherwise it's None

    def __repr__(self):
        return f"DurationCollector {self.order}"
    
    def __str__(self):
        return f"DurationCollector order={self.order} verbose={self.verbose} name={self.name} format={self.format}, source={self.source}"

    def get_base_type(self):
        return duration.Duration
    
    def process(self, key_durations, next_duration):
        index_str = MusicUtils.show_durations(key_durations)      # str value
        col_val = next_duration.duration.quarterLengthNoTuplets   # float value
        if self.verbose > 1:
            print(f"key: {index_str}, next: {col_val}")
        
        if self.counts_df is None:
            # initialize the counts DataFrame
            # self.counts_df = pd.DataFrame(data=[1],index=[index_str], columns=[str(next_duration.quarterLength)] )
            self.counts_df = pd.DataFrame(data=[1],index=[index_str], columns=[next_duration.duration.quarterLengthNoTuplets] )
        
        else:
            if index_str not in self.counts_df.index:   # add a new row
                self.counts_df.loc[index_str, col_val] = 1
            else: # update existing index
                if col_val in self.counts_df.columns:
                    self.counts_df.loc[index_str, col_val] = 1 + self.counts_df.loc[index_str, col_val]
                else:
                    self.counts_df.loc[index_str, col_val] = 1

        self.counts_df = self.counts_df.fillna(0)
        
    def set_source(self, source):
        """This method is called in the __init__ of the parent class, MusicCollector
        
        The source is notes DataFrame (notes_df) created by NoteCollector,
        or DataFrame (intervals_df) created by IntervalCollector
        """
        result = False
        if not isinstance(source, pd.DataFrame):
            # TODO: add support for stand-alone operation. 
            # TODO: Need to add DurationCollectorRunner and expand set_source() as in other MusicCollectors
            print("Invalid source", file=sys.stderr)
            return result
        self.source = source
        self.source_df = source
        self.durations_df = None
        result = True
        if self.source_df is not None:   # create the durations DataFrame
            self.durations_df = MusicUtils.get_durations_from_notes(self.source_df)
        
        return result

        
    def collect(self):
        """
        Run collection using the set parameters
        Returns MarkovChain result
        """
        if self.verbose > 1:
            print(f"durations: {self.durations_df}")
            
        if self.durations_df is not None:
            df_len = len(self.durations_df) 
            iloc = 0
            while iloc + self.order < df_len:
                key_durations = self.durations_df.iloc[iloc:iloc+self.order]    # list of length self.order
                next_duration = self.durations_df.iloc[iloc+self.order]
                self.process(key_durations, next_duration)      # add to counts DataFrame
                
                iloc = iloc + self.order

        self.counts_df.rename_axis('KEY', inplace=True)
        if self.sort_chain:
            self.counts_df.sort_index('index', ascending=True, inplace=True)
            self.counts_df.sort_index(axis=1, ascending=True, inplace=True)

        #
        # create the MarkovChain from the counts by added probabilities
        #
        sums = self.counts_df.sum(axis=1)
        self.chain_df = self.counts_df.div(sums, axis=0)
        self.chain_df.rename_axis('KEY', inplace=True)
        self.chain_df = self.chain_df.applymap(lambda x: MusicUtils.round_values(x, 3))
        
        self.markovChain = MarkovChain(self.order, self.counts_df,  chain_df=self.chain_df, myname=self.name)
        
        if self.verbose > 1:
            print(f" Counts:\n {self.counts_df}")
            print(f" MarkovChain:\n {self.markovChain}")
            if self.verbose > 1:
                print(self.markovChain.__repr__())

        return self.markovChain

    
    def save(self):
        """Saves the chain_df and counts_df DataFrames to a file in specified format.
        
        TODO - fix if order==1, the chain and counts index is saved as a float instead of a str.
        
        """
        save_result = super().save()
        if self.durations_df is not None and self.name is not None:
            #
            # optionally save durations_df as .csv
            #
            filename = "{}/{}_durations.{}".format(self.save_folder, self.name, self.format)
            df = self.durations_df[['type','ordinal','dots','fullName','quarterLength','tuplets']]
            df.to_csv(filename)
        return save_result
    
if __name__ == '__main__':
    print(DurationCollector.__doc__)
    


# ------------------------------------------------------------------------------
# Name:          producer.py
# Purpose:       PartsProducer class.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2022 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

import music
from common.producer import Producer
from music.musicUtils import MusicUtils
import pandas as pd
import random, math, sys
from music21 import note, clef, stream, interval, tempo, meter, key, metadata

class MusicProducer(Producer):
    """Produce music from interval, notes or scale degree MarkovChains.

    Known Issues: 
      Reading durations_chain in json format is incorrect. Currently the columns are interpreted as datetime values, should be int.
      MusicProducer intervals on existing chain file gets into an infinite loop. With the same parameters creating the chain files works fine.

    """

    def __init__(self, state_size, markovChain, durationsChain, source_file, parts, produceParts, collection_mode, num=10, verbose=0, rand_seed=42, producerType=None):
        super().__init__(state_size, markovChain, source_file, num=num, verbose=verbose,  rand_seed=rand_seed)
        
        self.instruments = music.Instruments(self.verbose)
        self.show = None
        self.producerType = producerType

        self.score = stream.Score()
        self.score.insert(0, metadata.Metadata())
        self.parts = parts                  # comma-delimited filter part names from the command line
        self.produceParts = produceParts    # comma-delimited production part names (instruments) from the command line
        self.part_names, self.part_numbers = MusicProducer.get_parts(parts)
        self.producePart_names = produceParts
        
        self.durationsChain = durationsChain
        self.collection_mode = collection_mode
        self.pitch_class_mode = (collection_mode == 'dpc' or collection_mode == 'apc')
        self.pitch_mode =  (collection_mode == 'dp' or collection_mode == 'ap')
        self.scale_degree_mode = collection_mode.startswith('sd')
        
        # if not None, this is the duration to use for all Notes instead of using durationsChain
        self.fixed_duration = None   # duration.Duration(quarterLength=0.5)
        
        self._set_keys()
        self.raw_seed = None
        self.score_key = key.Key('C')

        #
        # for producerType of 'intervals' there needs to be one start note per part
        #
        self.start_notes = dict()    # key is producePart name, value is a note.Note

        #
        # for producerType of 'notes' need to have the collection_mode (dp, dpc, ap, apc, sd)
        #
        self.collection_mode = collection_mode
        self.tempo = tempo.MetronomeMark(number=80, referent=note.Note(type='quarter'))
        self.timeSignature = meter.TimeSignature('4/4')
        self.num_measures = self.num        # number of measures to produce
        self.enforceRange = False           # force instruments to their range
        self.trace_mode = False             # displays the notes/intervals as they are produced

    
    @staticmethod
    def get_parts(parts:str) -> ([str],[int]):
        part_names = []
        part_numbers = []
        if parts is not None:
            parts_list = parts.split(",")   # could be numbers or names
            for p in parts_list:
                if p.isdigit():   # all digits
                    part_numbers.append(int(p))
                else:
                    part_names.append(p)
        return (part_names, part_numbers)
    

    def __str__(self):
        return f"MusicProducer order={self.order}, source={self.source_file}, num={self.num}, producerType={self.producerType}, \
collectionMode = {self.collection_mode}, parts={self.parts}, produce={self.produceParts}, verbose={self.verbose}"

    def add_part_notes(self, notes:str):
        notes_list = []
        if notes is None:
            if self.producerType == 'intervals':
                raise TypeError("You must specify a starting note for each part")
            else:
                # need n-starting notes where n is the number of producePart_names
                notes_list = self.get_seed(aseed=None)
        else:      
            notes_list = notes.split(",")
        if len(notes_list) < len(self.producePart_names):    # it's okay to have more notes than parts, just use what's there
            raise TypeError("You must specify a starting note for each part")
        
        for i in range(len(self.producePart_names)):
            self.start_notes[self.producePart_names[i]] = note.Note(notes_list[i])
    
    def _set_keys(self):
        initstr = None
        if self.producerType == 'intervals':
            initobj = music.IntervalCollector.initial_object
            initstr = '{}'.format(initobj.semitones)
        elif self.producerType == 'notes':
            if self.collection_mode == 'dp':
                initstr="C0"
            #
            # there actually isn't an initial key for dpc or sd collection mode
            # as there's no way in the current design to denote that
            #
            elif self.collection_mode == 'dpc':
                initstr = "C"
            elif self.collection_mode == 'sd':
                initstr = '1'
        
        self._keys = pd.Series(self.chain_df.index)
        self._key_values = self._keys.values
        self._initial_keys = self._keys[self._keys.apply(lambda x: x.startswith(initstr))]
        
        self._durationKeys = pd.Series(self.durationsChain.chain_df.index)
        self._durations_key_values = self._durationKeys.values

    def get_seed(self, aseed=None):
        theseed = self.seed
        if self.seed is None:
            # no seed already exists, so use the aseed argument if it's not None or pick one at random
            if aseed is not None:
                theseed = aseed
            elif self.initial:
                # pick a seed that starts a Part
                ind = random.randint(0, len(self.initial_keys)-1)
                theseed = self.initial_keys.iloc[ind]
            else:
                # pick any old seed from the chain's index
                ind = random.randint(0, len(self.keys)-1)
                theseed = self.keys.iloc[ind]
        return theseed
    
    def get_durations_seed(self):
            # pick any old seed from the durations chain's index
            return self._durationKeys[random.randint(0, len(self._durationKeys)-1)]
        
    def set_seed(self, raw_seed, collector_type):
        """Creates a seed for Notes or Intervals from a list
            Arguments:
                raw_seed - a [str] or [int] depending on collector_type
                collector_type - 'notes' or 'intervals'
            Returns:
                For collector_type of notes, the input is a [str], for example: ['A4', 'G4'],
                the result is a single string "A4,G4" in this case.
                For collector_type of intervals, the input is a [int], for example [-2, -1]
                the result is a single string, "-2,-1" in this case
        
        """
        if collector_type == "intervals":
            theseed = MusicUtils.to_string(raw_seed)
        elif collector_type == "notes":
            theseed = ",".join(raw_seed)
        
        self.seed = theseed
        return self.seed

    def get_next_seed(self):
        """
        Checks the seed_count against the recycle_seed_count
        and if >= gets a new seed, otherwise returns the existing one
        """
        self.seed_count = self.seed_count + 1
        aseed = self.seed
        if self.seed_count >= self.recycle_seed_count:
            # pick a new seed
            aseed = self.get_seed()
            self.seed_count = 0
        return aseed
    
    def get_next_object(self, seed):
        """Gets the next Note or Interval based on the seed argument
        
        Args:
            seed: The seed value to look up in the MarkovChain.
        
        Returns:
            Both the next Note or Interval and a new seed as a dict with keys 'new_seed' and 'next_token'
            The type of the next object returned is determined by self.type ('Notes' or 'Intervals')
        """
        
        new_seed = None
        next_token = None
        if self.trace_mode:
            print(f"get_next_object(seed): \"{seed}\"")
        if seed in self._key_values:
            row = self.chain_df.loc[seed]
            row_probs = row[row > 0].cumsum()
            # given a probability, pick the first entry in the row_probs (cumulative probabilities)
            # having a probability > the random probability
            # 
            prob = random.random()
            p = row_probs[row_probs> prob].iat[0]
            nt = row_probs[row_probs> prob].index[0]
            
            if self.producerType == 'intervals':    # intervals can only be integers (for now)
                ns = [int(x) for x in seed.split(',')]
                next_token = int(math.trunc(float(nt)))
            else:   # 'notes'
                ns = seed.split(',')
                next_token = nt
                
            ns.append(next_token)
            new_seed =  MusicUtils.to_string(ns[1:])
            if self.verbose > 1:
                print(f"random prob: {prob}, row prob: {p}, seed: '{seed}', next_token: '{next_token}', new_seed: '{new_seed}'")
        
        return dict([ ('next_token', next_token), ('new_seed',new_seed)])
    
    def get_next_durations_object(self, seed):
        """Gets the next Duration based on the seed argument
        
        Args:
            seed: The seed value to look up in the durations MarkovChain.
        
        Returns:
            Both the next duration and a new seed as a dict with keys 'new_seed' and 'next_token'
            The type of the next object returned is determined by self.type ('Notes' or 'Intervals')
            
            self.fixed_duration is not None, that is returned as 'next_token'
            and the seed argument is returned as 'new_seed'
            
        TODO - this doesn't work for order=1 chains, and it looks like it only works for order=2
        get_next_object probably has same problem.
        The problem is the durations chain keys for order 1 is a float instead of a string.
        
        """
        new_seed = None
        next_token = None
        if self.trace_mode:
            print(f"get_next_durations_object(seed): \"{seed}\"")
        if self.fixed_duration is not None:
            next_token = self.fixed_duration
            new_seed = seed
        else:
            if seed in self._durations_key_values:
                row = self.durationsChain.chain_df.loc[seed]
                row_probs = row[row > 0].cumsum()
                # given a probability, pick the first entry in the row_probs (cumulative probabilities)
                # having a probability > the random probability
                # 
                prob = random.random()
                p = row_probs[row_probs> prob].iat[0]
                next_token = float(row_probs[row_probs> prob].index[0])
                #
                # this next bit is needed because order=1 duration chain index (seed variable) is float instead of str
                # so it's a hack until that can be fixed
                #
                if isinstance(seed, float):
                    ns = seed   # a single float value
                    new_seed = next_token
                else:
                    ns = [float(x) for x in seed.split(',')]
                    ns.append(next_token)
                    new_seed = str(ns[1:]).replace(' ', '')
                    new_seed = new_seed[1:len(new_seed)-1]   # drop the []
                    
                if self.verbose > 1:
                    print(f"random prob: {prob}, row prob: {p}, seed: '{seed}', next_token: '{next_token}', new_seed: '{new_seed}'")
                    
        return dict([ ('next_token', next_token), ('new_seed',new_seed)])
    
    
    def produce(self):
        """Produce a Score from a MarkovChain
        
        """
        
        initial_seed = self.get_seed(self.seed)      # get the initial seed
        seed = initial_seed
        durations_seed = self.get_durations_seed()
        if self.verbose > 0:
            print(f"initial seeds: '{seed}' duration: {durations_seed}")
        if self.verbose > 2:
            print(f" MarkovChain:\n {self.markovChain}")
        #
        # partNotes_dict = dict(),  key is part name, value is [note.Note]
        #
        # total duration is sum of quarter lengths of all measures
        #
        total_duration = self.num_measures * self.timeSignature.numerator * self.timeSignature.beatDuration.quarterLength
        
        for pname in self.producePart_names:    # the part name must also be a valid Instrument
            if self.verbose > 0:
                print(f"---- Starting {pname} part")            
            part = stream.Part()
            clef = self.instruments.get_instrument_clef(pname)
            part_instrument = self.instruments.get_instrument(pname)

            part.insert(clef)
            part.insert(self.tempo)
            part.insert(part_instrument)
            part.insert(self.score_key)
            part.insert(self.timeSignature)
            part_notes = []
            
            if self.producerType == 'intervals':
                part_notes = self.produce_intervals(pname, seed, total_duration)
            elif self.producerType == 'notes':
                part_notes = self.produce_notes(pname, seed, total_duration)
            
            #
            # add rests to pad the last measure(s) of each part
            # so all parts have the same total duration (in terms of quarter lengths)
            # 
            if self.verbose > 0:
                print(f"---- End of {pname} part")
            part.append(part_notes)
            self.score.append(part)
        
        return self.score
    
    def produce_notes(self, part_name, seed, total_duration):
        """Produce a list of Note from a MarkovChain for 'notes' producerType for the named part

        """        
        part_notes = []
        more_to_go = True
        num_notes = 0
        part_duration = 0.0  # running total of the durations of all the part notes
        durations_seed = self.get_durations_seed()

        terminal_note = music.NoteCollector.terminal_object     # 'C#8'
            
        while more_to_go:       # more notes to add for this part
            next_object_dict = self.get_next_object(seed)
            next_duration_dict = self.get_next_durations_object(durations_seed)
            newnote_str = next_object_dict['next_token']
            #
            # for scale degrees, newnote_str is a scale degree
            # so need to convert to a note using the current Key
            #
            if self.collection_mode != 'sd':
                newnote = note.Note(newnote_str)
            else:
                newpitch = MusicUtils.get_pitch_from_scaleDegree(self.score_key, newnote_str)
                newnote = note.Note(newpitch)
                
            if self.trace_mode:
                print('newnote: {}\t new_seed: {}'.format(newnote.nameWithOctave, next_object_dict['new_seed']))
            if newnote == terminal_note:
                seed = self.get_seed()
                continue

            dur_str = next_duration_dict['next_token']    # quarterLength 
            if dur_str is None:
                durations_seed = self.get_durations_seed()
                continue
            dur = float(dur_str)
            if self.enforceRange and not self.instruments.is_in_range(part_name, newnote):
                #
                # if the note is out of range for this part (instrument)
                # transpose up/down the number of octaves needed to bring back into range
                #
                self.instruments.adjust_to_range(part_name, newnote, inPlace=True)

            remaining_dur =  total_duration - part_duration
            if part_duration + dur >= total_duration:
                dur = remaining_dur     # make the last note fill out the measure
                more_to_go = False
            if dur > 0:
                #
                # add rests to pad the last measure(s) of each part
                # so all parts have the same total duration (in terms of quarter lengths)
                #                 
                newnote.duration.quarterLength = dur
                if self.verbose > 1:
                    print(f"{newnote.fullName}")
                part_notes.append(newnote)

                part_duration = part_duration + dur
                seed =  next_object_dict['new_seed']
                if seed is None:
                    print("seed is None")
                durations_seed = next_duration_dict['new_seed']
                num_notes = num_notes + 1
                
        if self.verbose > 0:
            print(f"---- End of {part_name} part, duration: {part_duration}")
                    
        return part_notes
    
    def produce_intervals(self, part_name, seed, total_duration):
        """Produce list of Note from a MarkovChain for 'intervals' producerType for the named part

        """
        
        part_notes = []
        more_to_go = True
        num_notes = 1
        first_note = self.start_notes[part_name]
        part_notes.append(first_note)
        part_duration = first_note.duration.quarterLength       # running total of the durations of all the part notes
        durations_seed = self.get_durations_seed()
        prev_note = self.start_notes[part_name]
        
        while more_to_go:       # more notes to add for this part
            next_object_dict = self.get_next_object(seed)
            next_duration_dict = self.get_next_durations_object(durations_seed)
            if self.verbose > 1:
                print('next_token: {}\t new_seed: {}'.format(next_object_dict['next_token'], next_object_dict['new_seed']))
            #
            # create the next note
            #
            halfsteps = next_object_dict['next_token']
            if halfsteps == 100:
                #
                # if the next token is the terminal object (100 in this case)
                # pick a new seed and continue
                #
                seed = self.get_seed()
                continue
            ainterval = interval.Interval(halfsteps)
            if self.trace_mode:
                print(f"halfsteps: {halfsteps}, interval: {ainterval.name}")
            
            dur_str = next_duration_dict['next_token']    # quarterLength 
            if dur_str is None:
                durations_seed = self.get_durations_seed()
                continue
            dur = float(dur_str)
            if prev_note is None:
                print("Warning: prev_note is None", file=sys.stderr)
                continue
                
            newnote = prev_note.transpose(ainterval)
            if self.enforceRange and not self.instruments.is_in_range(part_name, newnote):
                #
                # if the note is out of range for this part (instrument)
                # transpose up/down the number of octaves needed to bring back into range
                #
                self.instruments.adjust_to_range(part_name, newnote, inPlace=True)

            remaining_dur =  total_duration - part_duration
            if part_duration + dur >= total_duration:
                dur = remaining_dur     # make the last note fill out the measure
                more_to_go = False
            if dur > 0:
                #
                # add rests to pad the last measure(s) of each part
                # so all parts have the same total duration (in terms of quarter lengths)
                #                 
                newnote.duration.quarterLength = dur
                if self.verbose > 1:
                    print(f"{newnote.fullName}")
                part_notes.append(newnote)

                part_duration = part_duration + dur
                prev_note = newnote
                seed =  next_object_dict['new_seed']
                durations_seed = next_duration_dict['new_seed']
                num_notes = num_notes + 1    

        if self.verbose > 0:
            print(f"---- End of {part_name} part, duration: {part_duration}")
    
        return part_notes
    
    
if __name__ == '__main__':
    print(MusicProducer.__doc__)
        
        



from __future__ import absolute_import

import sys
minPythonVersion = (3, 9)
minPythonVersionStr = '.'.join([str(x) for x in minPythonVersion])

del sys
del minPythonVersion
del minPythonVersionStr

__author__ = "Donald Bacon"
__version__ = "0.1.2"


# this defines what  is loaded when importing __all__
# put these in alphabetical order FIRST dirs then modules
# but: base must come first; in some cases other modules depend on
# definitions in base


__all__ = [
    # sub folders
    'common',
    'music',
    # individual modules
    # KEEP ALPHABETICAL unless necessary for load reasons, if so
    # put a note.  Keep one letter per line.
    'chain_test',
    'geo_test',
    'path_titles'
]

# -----------------------------------------------------------------------------
# this brings all of our own __all__ names into the dwbzen package namespace
# pylint: disable=wildcard-import
# 
import common
import music






# ------------------------------------------------------------------------------
# Name:          intervalCollector.py
# Purpose:       Interval collector class.
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2021 Donald Bacon

# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from music.musicCollector import MusicCollector
from music.musicUtils import MusicUtils
from common.markovChain import MarkovChain
from common.collectorProducer import CollectorProducer
import pandas as pd
from music21 import  interval, converter, corpus

class IntervalCollector(MusicCollector):
    """IntervalCollector creates a MarkovChains from a the Intervals in Score Notes.
    
    """
    
    terminal_object = interval.Interval(100)
    initial_object = interval.Interval(99)

    def __init__(self, state_size=2, verbose=0, source=None, parts=None):
        super().__init__(state_size, verbose, source, parts)
        
        self.initial_object, self.terminal_object = IntervalCollector.initialize_initial_terminal_objects()

        self.countsFileName = '_intervalsCounts' + '_0{}'.format(state_size)
        self.chainFileName = '_intervalsChain' + '_0{}'.format(state_size)
        if source is not None:
            self.source = self.set_source(source)   # otherwise it's None
    
    @staticmethod
    def initialize_initial_terminal_objects() -> (pd.DataFrame, pd.DataFrame):
        initial_dict = {'interval':IntervalCollector.initial_object, 'name':IntervalCollector.initial_object.name, 'directedName':IntervalCollector.initial_object.directedName,\
                        'niceName':IntervalCollector.initial_object.niceName, 'semitones':IntervalCollector.initial_object.semitones, 'part_number':1, 'part_name':'initial'}
        terminal_dict = {'interval':IntervalCollector.terminal_object, 'name':IntervalCollector.terminal_object.name, 'directedName':IntervalCollector.terminal_object.directedName,\
                        'niceName':IntervalCollector.terminal_object.niceName, 'semitones':IntervalCollector.terminal_object.semitones, 'part_number':1, 'part_name':'terminal'}
        initial_object = pd.DataFrame(data=initial_dict, index=[0]) 
        terminal_object = pd.DataFrame(data=terminal_dict, index=[0])
        return initial_object, terminal_object
        
    def __repr__(self):
        # TODO: this should return a serialized form 
        return f"IntervalCollector {self.order}"
    
    def __str__(self):
        return f"IntervalCollector order={self.order} verbose={self.verbose} name={self.name} format={self.format}, source={self.source}"

    def get_base_type(self):
        return interval.Interval
    
    def process(self, key_intervals, next_interval):
        index_str = MusicUtils.show_intervals(key_intervals,'semitones')
        col_str = str(next_interval.semitones)
        col_name = next_interval.interval.directedName
        
        if self.verbose > 1:
            print(f"key_interval: {index_str}, next_interval: {next_interval.semitones}, {col_name}")
    
        if self.counts_df is None:
            # initialize the counts DataFrame
            self.counts_df = pd.DataFrame(data=[1],index=[index_str], columns=[next_interval.semitones])
        else:
            if index_str not in self.counts_df.index:   # add a new row
                self.counts_df.loc[index_str, col_str] = 1
            else: # update existing index
                if col_str in self.counts_df.columns:
                    self.counts_df.loc[index_str, col_str] = 1 + self.counts_df.loc[index_str, col_str]
                else:
                    self.counts_df.loc[index_str, col_str] = 1
        self.counts_df = self.counts_df.fillna(0)

    def set_source(self, source):
        """Determine if source is a file or folder and if it exists (or not)
        
        Source can be a single file, or a composer name (such as 'bach')
        A single file must be compressed musicXML (.mxl), uncompressed (.musicxml), or .xml
        An xml file is treated as an uncompressed musicXML file
        A single file that resides in the music21 corpus is specified without an extension.
        The default corpus location ("/Compile/music21/music21/corpus") may be changed by setting corpus_folder attribute
        It may be accessed symbolically on the command line by the $CORPUS variable.
        Examples:
        --source composer='bach'                    # corpus search for composer 'bach'
        --source title='bwv*'                       # corpus wild card search
        --source "composer='bach',title='bwv*'"     # corpus wild card search for composer and title
        --source $CORPUS/haydn/opus74no1            # an individual corpus file
        --source '/data/music/Corpus/dwbzen/Prelude.mxl'    # a single musicXML file. The .mxl may be omitted
        --source [filename].json                    # load a serialized interval DataFrame (TODO)
        
        """
        result = True
        title = None
        composer = None
        if source.startswith('$CORPUS'):
            corpus_file = self.corpus_folder + source[6:]
            self.score = corpus.parse(corpus_file)
        elif 'composer' in source or 'title' in source:
            search_string = source.split(",")
            for ss in search_string:
                st = ss.split('=')
                if st[0] == 'composer':
                    composer = st[1]
                elif st[0] == 'title':
                    title = st[1]
            MusicUtils.verbose = self.verbose
            self.intervals_df, self.score_partNames, self.score_partNumbers = \
                MusicUtils.get_all_score_music21_objects(interval.Interval, composer=composer, title=title, partnames=self.part_names, partnumbers=self.part_numbers) 
            if self.intervals_df is None or len(self.intervals_df) == 0:
                result = False
        else:   # must be a single filename or path
            file_info = MusicUtils.get_file_info(source)
            if file_info['Path'].exists():
                self.score = converter.parse(file_info['path_text'])
                if self.verbose > 2:
                    print(self.score)
            if self.score is not None:
                self.intervals_df, self.score_partNames, self.score_partNumbers = \
                    MusicUtils.get_music21_objects_for_score(interval.Interval, self.score, self.part_names, self.part_numbers)
            else:
                result = False
        return result

    def collect(self) -> MarkovChain:
        """
        Run collection using the set parameters
        Returns MarkovChain result
        """
        if self.verbose > 1:
            print(f"intervals: {self.intervals_df}")
            
        for pname in self.score_partNames:
            partIntervals_df = self.intervals_df[self.intervals_df['part_name']==pname]
            df_len = len(partIntervals_df)
            key_intervals = None
            next_interval = None
            iloc = 0

            while iloc + self.order < df_len:
                if iloc == 0:
                    key_intervals =  self.initial_object.append(partIntervals_df[iloc:iloc+self.order-1])
                else:
                    key_intervals = partIntervals_df.iloc[iloc:iloc+self.order]    # list of length self.order
                    
                next_interval = partIntervals_df.iloc[iloc+self.order]
                self.process(key_intervals, next_interval)      # add to counts DataFrame
                
                iloc = iloc + 1
            
            # add the terminal_object to signify and of the part
            key_intervals = partIntervals_df.iloc[iloc:]
            next_interval = self.terminal_object.iloc[0]
            self.process(key_intervals, next_interval)

        self.counts_df.rename_axis('KEY', inplace=True)
        if self.sort_chain:
            self.counts_df.sort_index('index', ascending=True, inplace=True)
            self.counts_df.sort_index(axis=1, ascending=True, inplace=True)
        #
        # create the MarkovChain from the counts by summing probabilities
        #
        sums = self.counts_df.sum(axis=1)
        self.chain_df = self.counts_df.div(sums, axis=0)
        self.chain_df.rename_axis('KEY', inplace=True)
        self.chain_df = self.chain_df.applymap(lambda x: MusicUtils.round_values(x, 6))
        
        self.markovChain = MarkovChain(self.order, self.counts_df,  chain_df=self.chain_df, myname=self.name)
        
        if self.verbose > 1:
            print(f" Counts:\n {self.counts_df}")
            print(f" MarkovChain:\n {self.markovChain}")
            if self.verbose > 1:
                print(self.markovChain.__repr__())
        #
        # collect durations from the Score and notes_df DataFrame
        #
        self.collect_durations(self.intervals_df)
        return self.markovChain
    
    def save(self):
        """Saves the chain_df, counts_df and intervals_df DataFrames to a file in specified format.
        
        """

        save_result = super().save()
        if self.intervals_df is not None:
            #
            # optionally save intervals_df as .csv
            #
            filename = "{}/{}_intervals.{}".format(self.save_folder, self.name, self.format)
            df = self.intervals_df[['name','niceName','semitones','part_name','part_number']]
            save_result = CollectorProducer.save_df(df, filename, self.format, orient='records')
        return save_result
        
if __name__ == '__main__':
    print(IntervalCollector.__doc__)
        



'''
Created on Jul 9, 2021

@author: don_bacon
'''
import markovify

if __name__ == '__main__':
    with open("C:/data/text/ferlinghetti.txt") as f:
        text = f.read()
    
    # Build the model.
    text_model = markovify.Text(text, state_size=2)
    
    # Print five randomly-generated sentences
    for i in range(5):
        print(f'{i}: {text_model.make_sentence()}' )
    
    # Print three randomly-generated sentences of no more than 280 characters
    for i in range(3):
        print(f'{i}  {text_model.make_short_sentence(280)}' )
    
    json_text = text_model.to_json()
    print(f'json: {json_text}')
    